# Using both qualitative and quantitative data to improve parameter identification
# Eshan D. Mitra, Raquel Dias, Richard G. Posner, William S. Hlavacek

# Supplementary file S2.: Python implementation of the model of Raf dimerization and inhibition.
# The functions included in this file can be used to perform the analyses presented in the study.
# Software compatibility: Python 3.6, sympy 1.0, numpy 1.13.3, scipy 1.0.0

# --- SECTION 0: INITIAL SETUP ---

# Load sympy and define variables
import sympy as sp
import numpy as np
import matplotlib.pyplot as plt
import pickle
import scipy.optimize as optimization
from scipy.integrate import odeint
import os.path

# Function to generate NiceTable50.bp containing the ground truth data, if we don't already have it
def data_generator():
    def generate_data2(ind_var_1, ind_var_2, point_list, dep_var_dict, nsolve=True, guess_list=None):
        """
        Generates a custom data set
        2 Independent variables are ind_var_1 and ind_var_2, expressed in terms of R, I, RR, RI, RIR, and RIRI
        Returns the value of the dependent variable dep_var, expressed in terms of the same.

        :param ind_var_1: sympy expression for the first independent variable
        :param ind_var_2: sympy expression for the 2nd independent variable
        :param point_list: list of ordered pairs [(iv1,iv2)] to evaluate
        :param dep_var: sympy expression for the dependent variable
        :param nsolve: Should we use sympy.nsolve rather than a symbolic solve?
        :return:
        """

        if guess_list == None:
            guess_list = [2.] * 6

        IV1, IV2 = sp.symbols('IV1 IV2')
        eqs_iv = [ind_var_1 - IV1, ind_var_2 - IV2]
        # eqs_all = eqs_k + eqs_iv

        dep_var_funcs = {k: sp.lambdify((R, I, RR, RI, RIR, RIRI), dep_var_dict[k]) for k in dep_var_dict}

        if nsolve:
            if guess_list is None:
                guess_list = [2.] * len(point_list)
            result_list = dict()
            eqs_k_numeric = [e.subs(ground_const) for e in eqs_k]
            for iv1i, iv2i in point_list:
                eqs_iv_numeric = [e.subs({IV1: iv1i, IV2: iv2i}) for e in eqs_iv]
                # try:
                nsol_matrix = sp.nsolve(eqs_k_numeric + eqs_iv_numeric, species, guess_list)
                nsols = {species[i]: float(nsol_matrix[i]) for i in range(len(species))}
                result_list[(iv1i, iv2i)] = {k: dep_var_funcs[k](nsols[R], nsols[I], nsols[RR],
                                                                 nsols[RI], nsols[RIR], nsols[RIRI]) for k in dep_var_funcs}
                # except:
                #     # nsolve failed
                #     result_list[(iv1i,iv2i)] = None
        else:
            raise NotImplementedError

        return result_list



    R, RR, RI, RIR, RIRI, I, Rtot, Itot = sp.symbols('R RR RI RIR RIRI I Rtot Itot')
    K1, K2, K3, K5 = sp.symbols('K1 K2 K3 K5')
    species = [R, RR, RI, RIR, RIRI, I]

    # Define the known equations of the model
    eqs_k = [K1-RR/R**2, K2-RI/(R*I), K3-RIR/(RR*I), K5-RIRI/(RIR*I)]

    # Solve the system of equations symbolically
    # Here, the solutions are a positive and negative square root, we manually select the positive one.
    sols_all_R_I = sp.solve(eqs_k,[RR,RI,RIR,RIRI])
    sols_all_R_I

    # Turn the sols_all into callable functions of R and I, with the K1-K5 already plugged in.
    # This is needed for high-throughput generation of ground_data because sp.subs() is too slow.

    funcs_R_I = {k:sp.lambdify((R,I),sols_all_R_I[k].subs(ground_const)) for k in sols_all_R_I}
    funcs_R_I[RR](5,42)

    # Choose data points (RIR, RIRI) for which we should calculate ground truth data.

    # Try sampling log-distributed points instead
    # therange = np.linspace(-6.907,3.912,500) #0.001 to 50 on a log scale
    theRrange = np.linspace(np.log(0.003),np.log(20),500)
    theIrange = np.linspace(np.log(4e-5),np.log(300),500)
    rlist = [np.exp(i) for i in theRrange]
    ilist = [np.exp(i) for i in theIrange]

    points = [(ar, ai) for ar in rlist for ai in ilist]

    # Create the ground truth data

    # Apply the callable functions to each point in points, to make data set ground_data
    ground_data = {(ar,ai):{R:ar, I:ai, **{k:funcs_R_I[k](ar,ai) for k in funcs_R_I}} for (ar,ai) in points}

    Ybar_expr = (RI+RIR+2*RIRI)/(R+2*(RR+RIR+RIRI)+RI)
    myvariables = {'R':R,'I':I,'RI':RI,'RIR':RIR,'RR':RR,'RIRI':RIRI,'Ybar':Ybar_expr,'A':RR+RIR,'D':RR+RIR+RIRI,
                   'Itot':I+RI+RIR+2*RIRI}

    Rtot_expr = R+RI+2*(RR+RIR+RIRI)
    Ybar_expr = (RI+RIR+2*RIRI)/(R+2*(RR+RIR+RIRI)+RI)
    target_Rtot = 50.

    Rtot_func = sp.lambdify((R,RI,RR,RIR,RIRI),Rtot_expr)

    extracted_Ybar_data = dict()

    failures = []

    #Make trivial funcs for R and I
    funcs_R_I = {**funcs_R_I,R:sp.lambdify((R,I),R),I:sp.lambdify((R,I),I)}

    for k in ground_data:
        check = Rtot_func(ground_data[k][R],ground_data[k][RI],ground_data[k][RR],ground_data[k][RIR],
                          ground_data[k][RIRI])
        if check/target_Rtot > 0.95 and check/target_Rtot < 1.05:
            guess = [funcs_R_I[s](k[0],k[1]) for s in species]

            try:
                newpoint = generate_data2(Rtot_expr,I,[(target_Rtot,ground_data[k][I])],myvariables,guess_list=guess)
            except:
                failures.append(k)
                continue
            extracted_Ybar_data = {**extracted_Ybar_data, **newpoint}

    # Save the data
    import pickle
    f = open('NiceTable50.bp','wb')
    pickle.dump(extracted_Ybar_data,f)
    f.close()


# Font size incantation
plt.rc('font', **{'size': 16})
cbm = ['#009292', '#db6d00', '#b66dff', '#920000', '#24ff24', '#ff6db6']

# Initialize the sympy variables
R, RR, RI, RIR, RIRI, I, Rtot, Itot = sp.symbols('R RR RI RIR RIRI I Rtot Itot')
K1, K2, K3, K5 = sp.symbols('K1 K2 K3 K5')
species = [R, RR, RI, RIR, RIRI, I]

# Set the ground truth parameters to solve for.
ground_const = {K1: 0.04, K2: 20., K3: 4000., K5: 0.1}

# We assume that these values are known already
oracle_vals = {K1: 0.04, K2: 20., Rtot: 50.}

# Generate the data file if it was not generated on a previous run.
if not os.path.isfile('NiceTable50.bp'):
    data_generator()

# Reload the data that was acquired
f = open('NiceTable50.bp', 'rb')
all_data_50 = pickle.load(f)
f.close()


# --- SECTION 1: DATA GENERATION ---

def view_ground_data(which='All'):
    """
    Shows graphs of the ground truth data, which was loaded from NiceTable50.bp, generated from the Jupyter notebook

    :param: which - Which ground truth data graph to show, options are 'All', 'Ybar', 'ItotA', 'IA'
    :return:
    """
    if which == 'Ybar' or which == 'All':
        data = [(np.log10(key[1]), all_data_50[key]['Ybar']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        plt.figure()
        plt.scatter(xlist, ylist, alpha=0.01)
        plt.xlabel('log10([I])')
        plt.ylabel('Proportion of bound R')
        plt.tight_layout()

    if which == 'ItotA' or which == 'All':
        data = [(np.log10(all_data_50[key]['Itot']), all_data_50[key]['A']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        plt.figure()
        plt.scatter(xlist, ylist, alpha=0.01)
        plt.xlabel('log([Itot])')
        plt.ylabel('Activity')
        plt.tight_layout()

    if which == 'IA' or which == 'All':
        data = [(np.log10(all_data_50[key]['I']), all_data_50[key]['A']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        plt.figure()
        plt.scatter(xlist, ylist, alpha=0.01)
        plt.xlabel('log([I])')
        plt.ylabel('Activity')
        plt.tight_layout()


def expt_data(ground, ind_var, points, dep_var, noise, qualitative=False, control_value=None):
    """
    Generate experimental data; a list of (ind_var,dep_var) pairs,
    based on the ground data table ground.
    ground is assumed to be at constant Rtot (or some other independent, unvaried variable). all_data_50 is a good thing
    to pass for this.


    :param ground: The ground truth data table, such as is saved in NiceTable50. Assumed to be a table of constant
    Rtot (or whatever independent variable is to be held constant)
    :param ind_var:  The str key in ground corresponding to the independent variable to be varied.
    :param points: An iterable of the ind_var values to use.
    :param dep_var: The str key in ground corresponding to the dependent variable
    :param noise: The standard deviation of gaussian noise to add to the dependent variable.
    For qualitative data, this is instead the sensitivity - if the difference is smaller than this, no result.
    :param qualitative: If true, the resulting data is qualitative only. Returned dep_var values are just -1 (lower)
    0 (not significant), or 1 (higher)
    :param control_value: For qualitative mode, the dep_var value to compare to.
    :return: (x,y) pairs corresponding to the experimental data.
    """

    ground_keys = ground.keys()
    ground_vals = [ground[k] for k in ground_keys]

    iv_vals = [x[ind_var] for x in ground_vals]
    dv_vals = [x[dep_var] for x in ground_vals]

    return_data = []

    for p in points:
        closest_index = min(range(len(iv_vals)), key=lambda i: abs(iv_vals[i] - p))
        iv_i = iv_vals[closest_index]
        if qualitative:
            dv_i = dv_vals[closest_index]
            if abs(dv_i - control_value) < noise:
                dv_q = 0
            elif dv_i < control_value:
                dv_q = -1
            else:
                dv_q = 1
            return_data.append((iv_i, dv_q))

        else:
            dv_i = dv_vals[closest_index] + np.random.normal(0., noise)
            return_data.append((iv_i, dv_i))
    return return_data


def gen_expt_data_set(which, plot_flag=False, seed=413, n=13):
    """
    Generate a certain pre-defined experimental data set.

    Choices are:
    Ybar - Quantitative I vs Ybar data
    ItotA - Qualitative Itot vs A data
    IA - Qualitative I vs A data

    :param which: str specifying one of the choices above
    :param plot_flag: If true, show a plot of the data.
    :param seed: Seed for random number generator (quantitative data only)
    :param n: Number of data points to generate
    :return:
    """

    if which == 'Ybar':
        mypoints = [10. ** (x) for x in np.linspace(-4., 2, n)]
        np.random.seed(seed)
        Ybar_expt = expt_data(all_data_50, 'I', mypoints, 'Ybar', 0.1)
        if plot_flag:
            plt.figure()
            plt.scatter([np.log10(p[0]) for p in Ybar_expt], [p[1] for p in Ybar_expt])
            plt.xlabel('Log(I)')
            plt.ylabel('Ybar')
        return Ybar_expt

    if which == 'ItotA' or which == 'IA':
        # Solve for the activity value with no inhibitor, to be used as a standard for the qualitative data.
        # Note Rtot = 50 is hard-coded here.
        activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - ground_const[K1]], [R, RR])
        activity0 = activity0_sols[1][1]
        print(activity0)

        if which == 'ItotA':
            mypoints = [10 ** x for x in np.linspace(0.5, 2.3, n)]
            A_qual = expt_data(all_data_50, 'Itot', mypoints, 'A', 1., True, activity0)
            if plot_flag:
                plt.figure()
                plt.scatter([np.log10(p[0]) for p in A_qual], [p[1] for p in A_qual])
                plt.xlabel('Log(Itot)')
                plt.ylabel('Qualitative Activity')
            return A_qual

        if which == 'IA':
            mypoints = [10. ** (x) for x in np.linspace(-4., 2, n)]
            A_qual_I = expt_data(all_data_50, 'I', mypoints, 'A', 1.5, True, activity0)

            if plot_flag:
                plt.scatter([np.log10(p[0]) for p in A_qual_I], [p[1] for p in A_qual_I])
                plt.ylabel('Qualitative Activity')
                plt.xlabel('Log(I)')
            return A_qual_I

    raise ValueError('Invalid option')


# ---- SECTION 2: THE OPTIMIZER CLASS ----

class Optimizer:
    def __init__(self, quant, qual, quant_data=None, quant_func=None, qual_data=None, qual_func=None, qual_0=None,
                 qual_c=100., sensitivity=1.5):
        """
        Create an Optimizer for quantiative and/or qualitative optimization.

        If you are only doing one of them (quantitative or qualitative), you may omit the arguments associated with
        the other one.

        :param quant: If true, the fitting includes quantitative data
        :param qual: If true, the fitting includes qualitative data
        :param quant_data: quantitative data list, as returned by expt_data(), a list of (<ind_var>,<dep_var>)
        :param quant_func: A function of (<ind_var>, Rtot, K3, K5) that returns the value of <dep_var> for the
        quantitative data. [warning: order of args changed compared to Jupyter notebook]
        :param qual_data: qualitative data list, as returned by expt_data()
        :param qual_func: A function, of the same form as quant_func
        :param qual_0: Baseline value that the qualitative data is based on (e.g. activity0)
        :param qual_c: Constant that multiplies the penalty function.
        :param sensitivity: Sensitivity of qualitative data == allowable error for equality constraints
        """
        self.quant = quant
        self.qual = qual
        self.quant_data = quant_data
        self.quant_func = quant_func
        self.qual_data = qual_data
        self.qual_func = qual_func
        self.qual_0 = qual_0
        self.qual_c = qual_c
        self.sensitivity = sensitivity

    @staticmethod
    def target(x, self):
        # target function to minimize
        # Takes self as the 2nd param so that optimization can use the 1st param.
        return self.objective(x[0], x[1]) + self.penalty(x[0], x[1])

    @staticmethod
    def target_const_k3(x, self, k3):
        # target function to minimize (for profile likelihood with constant K3)
        # Takes self as the 2nd param so that optimization can use the 1st param.
        return self.objective(k3,x[0]) + self.penalty(k3, x[0])

    @staticmethod
    def target_const_k5(x, self, k5):
        # target function to minimize (for profile likelihood with constant K5)
        # Takes self as the 2nd param so that optimization can use the 1st param.
        return self.objective(x[0], k5) + self.penalty(x[0], k5)

    def objective(self, pK3, pK5):

        if self.quant:
            # Least squares fit to the quantitative data.
            return sum([(self.quant_func(p[0], oracle_vals[Rtot], pK3, pK5) - p[1]) ** 2 for p in self.quant_data])
        else:
            # No quantitative contribution to the objective
            return 0.

    def penalty(self, pK3, pK5, beta=1.):
        # return sum([max(0.,c(pK3,pK5))**beta for c in constraint_list])
        # The constraint_list approach seems impossible because of how lambda functions work.
        # Have to instead piece through the data list in one shot here.
        if self.qual:
            total = 0.
            for p in self.qual_data:
                if p[1] == 0:
                    # Do an equality constraint. Must lie within sensitivity of activity0
                    constraint_val = (self.qual_func(p[0], oracle_vals[Rtot], pK3, pK5) - self.qual_0)
                    if abs(constraint_val) > self.sensitivity:
                        total += abs(constraint_val)
                else:
                    constraint_val = (self.qual_func(p[0], oracle_vals[Rtot], pK3, pK5) - self.qual_0) * -p[1]
                    total += max(0., constraint_val) ** beta
            return total * self.qual_c
        else:
            return 0.

    def run(self):

        #answer = optimization.differential_evolution(Optimizer.target, [(1e-6, 1e6), (1e-6, 1e6)], maxiter=1000,
        #                                             seed=413, args=(self,))
        answer = optimization.differential_evolution(Optimizer.target, [(1e-4, 1e4), (1e-4, 1e4)], maxiter=1000,
                                                    seed=914, args=(self,), strategy='best1exp')
        return answer

    def profile_likelihood(self, const_param, points):

        if const_param == 'K3' or const_param == K3:
            target = Optimizer.target_const_k3
        elif const_param == 'K5' or const_param == K5:
            target = Optimizer.target_const_k5
        else:
            raise ValueError('The const_param must be either K3 or K5')

        results = []
        np.random.seed(914)
        for k3i in points:
            answer = optimization.differential_evolution(target, [(1e-4, 1e4)], maxiter=1000, args=(self, k3i),
                                                         strategy='best1exp')
            results.append(answer.fun)

        return results


# ---- SECTION 3: FITTING RUNS ----
# Run various preset fitting runs

Ybar_func = None


def get_Ybar_func():
    """
    Generates a callable function Ybar = f(I, Rtot, K3, K5) using algebraic manipulation in sympy

    The first time we call this, we compute it (slow), and save it as a global. Subsequent calls just return
    that global.
    :return:
    """
    global Ybar_func

    if Ybar_func:
        return Ybar_func

    # Define the known equations of the model
    eqs_k = [K1 - RR / R ** 2, K2 - RI / (R * I), K3 - RIR / (RR * I), K5 - RIRI / (RIR * I)]
    eqs_tot = [R + RI + 2 * (RR + RIR + RIRI) - Rtot, I + RI + RIR + 2 * RIRI - Itot]

    # Solve for Ybar algebraically
    Ybar = sp.symbols('Ybar')
    eq_Ybar = (RI + RIR + 2 * RIRI) / (R + 2 * (RR + RIR + RIRI) + RI) - Ybar
    sols = sp.solve(eqs_k + [eq_Ybar], [Ybar, RI, RR, RIR, RIRI])
    Ybar_expr = sols[0][0].subs(oracle_vals)

    # Solve for Rtot algebraically
    eq_Rtot = R + RI + 2 * (RR + RIR + RIRI) - Rtot
    sols = sp.solve(eqs_k + [eq_Rtot], [Rtot, RI, RR, RIR, RIRI])
    Rtot_expr = sols[Rtot].subs(oracle_vals)

    # Invert the Rtot function to get a function for R
    # Note that we checked by hand - solution[0] is the positive solution
    Rsols = sp.solve([Rtot_expr - Rtot], [R])

    final_eq = Ybar_expr.subs({R: Rsols[0][0]})
    final_function = final_func = sp.lambdify((I, Rtot, K3, K5), final_eq)

    Ybar_func = final_function
    return final_function


A_func = None


def get_A_func():
    """
    Generates a callable function A = f(I, Rtot, K3, K5) using algebraic manipulation in sympy.

    The first time we call this, we compute it (slow), and save it as a global. Subsequent calls just return
    that global.
    :return:
    """
    global A_func
    if A_func:
        return A_func

    eqs_k = [K1 - RR / R ** 2, K2 - RI / (R * I), K3 - RIR / (RR * I), K5 - RIRI / (RIR * I)]
    eqs_tot = [R + RI + 2 * (RR + RIR + RIRI) - Rtot, I + RI + RIR + 2 * RIRI - Itot]

    A = sp.symbols('A')
    eq_A = RR + RIR - A
    sols = sp.solve(eqs_k + [eq_A], [A, RI, RR, RIR, RIRI])
    A_expr = sols[A].subs(oracle_vals)

    # Solve for Rtot algebraically
    eq_Rtot = R + RI + 2 * (RR + RIR + RIRI) - Rtot
    sols = sp.solve(eqs_k + [eq_Rtot], [Rtot, RI, RR, RIR, RIRI])
    Rtot_expr = sols[Rtot].subs(oracle_vals)

    # Invert the Rtot function to get a function for R
    Rsols = sp.solve([Rtot_expr - Rtot], [R])

    final_A_eq = A_expr.subs({R: Rsols[0][0]})
    final_A_func = sp.lambdify((I, Rtot, K3, K5), final_A_eq)

    A_func = final_A_func
    return final_A_func


def ode_Itot_A_func(aItot, aRtot, aK3, aK5):
    """
    Computes the value of A based on parameters Itot, Rtot, K3, and K5

    :return:
    """
    sim_run = simulate_ode(aItot,aRtot,aK3,aK5)
    return sim_run[-1, 3]+sim_run[-1, 4]


def simulate_ode(itot, rtot, k3, k5):
    return odeint(raf_odes, (rtot, itot, 0., 0., 0., 0.), t=np.linspace(0,5,10000), args=(k3, k5))


def raf_odes(ys, t, k3, k5):
    k1 = oracle_vals[K1]
    k2 = oracle_vals[K2]
    k4 = k1 * k3 / k2
    k6 = k1 * k3 * k5 / k2 ** 2
    (r, i, ri, rr, rir, riri) = ys
    result = (
        +(-2) * 0.5 * 2 * k1 * r * r - k2 * r * i + 2 * 1 * rr + 1 * ri - k4 * r * ri + 1 * rir,
        - k2 * r * i + 1 * ri - k3 * rr * i + 1 * rir - k5 * rir * i + 1 * riri,
        + k2 * r * i - 1 * ri - k4 * r * ri + (-2) * 0.5 * 2 * k6 * ri * ri + 1 * rir + 2 * 1 * riri,
        + 0.5 * 2 * k1 * r * r - 1 * rr - k3 * rr * i + 1 * rir,
        + k3 * rr * i + k4 * r * ri - 1 * rir - 1 * rir - k5 * rir * i + 1 * riri,
        + 0.5 * 2 * k6 * ri * ri + k5 * rir * i - 1 * riri - 1 * riri

    )
    return result

def plot_Itot_A_func():
    """
    Makes a graph of Itot vs A to show that the Itot_A_func is working correctly

    Validated against view_ground_data() from NiceTable50

    For the ODE function, some points have small inaccuracies.
    :return:
    """
    Itots = np.linspace(-1, 2.5, 43)
    As = [ode_Itot_A_func(10. ** i, 50., 4000, 0.1) for i in Itots]
    plt.plot(Itots, As)
    plt.xlabel('Itot')
    plt.ylabel('Activity')


def quant_fit(plot_flag=False, plot_combined_fit=True):
    """
    Run a fit on quantitative I vs Ybar data!
    :return:
    """
    Ybar_data = gen_expt_data_set('Ybar')
    Ybar_function = get_Ybar_func()
    opt = Optimizer(True, False, quant_data=Ybar_data, quant_func=Ybar_function)
    fit = opt.run()
    if plot_flag:
        if plot_combined_fit:
            plotfit = tot_fit()
        else:
            plotfit = fit
        plt.figure(figsize=(4,3))
        kvals = plotfit.x
        data = [(np.log10(all_data_50[key]['I']), all_data_50[key]['Ybar']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        plt.plot(xlist, ylist, label='Ground truth', linewidth=2, color=cbm[5])
        plt.scatter([np.log10(p[0]) for p in Ybar_data], [p[1] for p in Ybar_data], marker='d', color=cbm[3], label='Data')

        xfit = np.linspace(-4, 2, 31)
        yfit = [Ybar_function(10 ** x, 50, kvals[0], kvals[1]) for x in xfit]
        plt.plot(xfit, yfit, color=cbm[0], label='Fit', linewidth=2)
        plt.xlabel("$\\mathregular{log_{10}}$(I)")
        plt.ylabel('$\\overline{Y}$')
        plt.tight_layout()
        # plt.legend(fontsize=12)
    return fit


def qual_fit(plot_flag=False, plot_combined_fit=True):
    # activity0 can be obtained from the oracle_vals, computing RR @ zero I
    activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - oracle_vals[K1]], [R, RR])
    activity0 = float(activity0_sols[1][1]) # float() required because ylim complains about type
    A_data = gen_expt_data_set('IA')
    A_func = get_A_func()
    opt = Optimizer(False, True, qual_data=A_data, qual_func=A_func, qual_0=activity0, sensitivity=1.5)
    fit = opt.run()
    if plot_flag:
        if plot_combined_fit:
            plotfit = tot_fit()
        else:
            plotfit = fit

        plt.figure(figsize=(4,3))
        kvals = plotfit.x

        data = [(np.log10(all_data_50[key]['I']), all_data_50[key]['A']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        # plt.scatter(xlist, ylist, alpha=1.0, color='#000000', label='Ground truth')
        plt.plot(xlist, ylist, label='Ground truth', linewidth=2, color=cbm[5])

        plt.scatter([], [], color='w', label='Data') # Just for the legend entry

        xfit = np.linspace(-4, 2, 31)
        yfit = [A_func(10 ** x, 50, kvals[0], kvals[1]) for x in xfit]
        plt.plot(xfit, yfit, color=cbm[0], label='Fit', linewidth=2)

        plt.axhline(y=activity0, color='k')
        plt.axhline(y=activity0-1.5, color='k', linestyle='--')
        plt.axhline(y=activity0+1.5, color='k', linestyle='--')

        #plt.xlabel('$log_{10}(I)$')
        #plt.xlabel('Log(I)')
        plt.xlabel("$\\mathregular{log_{10}}$(I)")
        plt.ylabel('Activity') #, color='r')
        print(activity0)
        plt.ylim([-1.5,28])

        # ax2 = plt.gca().twinx()
        #
        # ax2.scatter([np.log10(p[0]) for p in A_data], [activity0 + p[1]*activity0 for p in A_data], color='g', label='Data')
        # plt.ylim([-1,2*activity0+1])
        # plt.ylabel('Qualitative Activity', color='g')
        # plt.yticks((0,activity0,2*activity0), ('-', '0', '+'))
        for p in A_data:
            plt.legend(fontsize=12, loc=(-4, 3), bbox_to_anchor=(0, 0, 1, 1))
            if p[1] == -1:
                color = cbm[3]
                sym = u'$\u2013$' # en dash
                yy = 0.2 # Compensate for the stupid feature that not all characters print at the right vert alignment
                textcolor = 'w'
            elif p[1] == 1:
                color = cbm[2]
                sym = r'$+$'
                yy = -0.2
                textcolor = 'w'
            else:
                color = '#ffffff'
                sym = r'$0$'
                yy = -0.2
                textcolor = 'k'
            plt.scatter([np.log10(p[0])], [0], marker='o', c=color, s=130, linewidths=1,
                        edgecolors='k')
            # plt.scatter([np.log10(p[0])], [0], marker=sym, c='k', s=50)
            plt.text(np.log10(p[0]), yy, sym, fontname='STIXGeneral', size=10, va='center', ha='center', clip_on=True,
                     color=textcolor, weight='bold')

        plt.tight_layout()
        # plt.legend(fontsize=12, loc=(0.1, 0.1))

    return fit

def tot_fit(qual_c=0.03):
    activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - oracle_vals[K1]], [R, RR])
    activity0 = float(activity0_sols[1][1])  # float() required because ylim complains about type
    A_data = gen_expt_data_set('IA')
    A_func = get_A_func()

    Ybar_data = gen_expt_data_set('Ybar')
    Ybar_function = get_Ybar_func()
    opt = Optimizer(True, True, qual_data=A_data, qual_func=A_func, qual_0=activity0, sensitivity=1.5,
                    quant_data=Ybar_data, quant_func=Ybar_function, qual_c=qual_c)
    fit = opt.run()
    return fit


def qual_fit_Itot(plot_flag=False):
    # activity0 can be obtained from the oracle_vals, computing RR @ zero I
    activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - oracle_vals[K1]], [R, RR])
    activity0 = float(activity0_sols[1][1])
    A_data = gen_expt_data_set('ItotA')

    opt = Optimizer(False, True, qual_data=A_data, qual_func=ode_Itot_A_func, qual_0=activity0, sensitivity=1.)
    fit = opt.run()
    if plot_flag:
        plt.figure()
        kvals = fit.x


        data = [(np.log10(all_data_50[key]['Itot']), all_data_50[key]['A']) for key in all_data_50]
        [xlist, ylist] = zip(*data)
        plt.plot(xlist, ylist, linewidth=2, label='Ground truth')

        plt.scatter([],[], color='g', label='Data')


        xfit = np.linspace(0, 2.5, 31)
        yfit = [ode_Itot_A_func(10 ** x, 50, kvals[0], kvals[1]) for x in xfit]
        plt.plot(xfit, yfit, color='r', label='Fit')

        plt.legend()


        plt.axhline(y=activity0, color='k')
        plt.axhline(y=activity0 - 1., color='k', linestyle='--')
        plt.axhline(y=activity0 + 1., color='k', linestyle='--')

        plt.xlabel('Log(Itot)')
        plt.ylabel('Activity', color='r')

        plt.ylim([-1, 2 * activity0 + 1])

        ax2 = plt.gca().twinx()

        ax2.scatter([np.log10(p[0]) for p in A_data], [activity0 + p[1] * activity0 for p in A_data], color='g',
                    label='Data')
        plt.ylim([-1, 2 * activity0 + 1])
        plt.ylabel('Qualitative Activity', color='g')
        plt.yticks((0, activity0, 2 * activity0), ('-', '0', '+'))

        plt.tight_layout()

    return fit


def profiles(const_param,which,qual_c=0.03, pointcount=25, savestem=None, quantpts=13, qualpts=13):
    """

    :param const_param: Which parameter to hold constant, K3 or K5
    :param which: Which data set to use: quant, qual, or both
    :param qual_c: The qualitative C value. 0.03 is a nice compromise to count both kinds of data.
    :return:
    """

    if const_param=='K3':
        points = [10 ** x for x in np.linspace(2, 5, pointcount)]
    elif const_param=='K5':
        points = [10 ** x for x in np.linspace(-3, 0, pointcount)]

    plt.figure(figsize=(4,4))

    if which=='quant' or which=='all' or which=='alltot':
        Ybar_data = gen_expt_data_set('Ybar', n=quantpts)
        Ybar_function = get_Ybar_func()
        opt = Optimizer(True, False, quant_data=Ybar_data, quant_func=Ybar_function)
        profile = opt.profile_likelihood(const_param, points)
        plt.plot([np.log10(x) for x in points], profile, 'o-', label='Quantitative', color=cbm[0])
        if savestem:
            f = open(savestem+'_quant', 'wb')
            pickle.dump(([np.log10(x) for x in points], profile), f)
            f.close()

    if which=='qual' or which=='all':
        activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - oracle_vals[K1]], [R, RR])
        activity0 = activity0_sols[1][1]
        A_data = gen_expt_data_set('IA', n=qualpts)
        A_func = get_A_func()
        opt = Optimizer(False, True, qual_data=A_data, qual_func=A_func, qual_0=activity0, sensitivity=1.5,
                        qual_c=qual_c)
        profile = opt.profile_likelihood(const_param, points)
        plt.plot([np.log10(x) for x in points], profile, 's-', label='Qualitative', color=cbm[1])
        if savestem:
            f = open(savestem+'_qual', 'wb')
            pickle.dump(([np.log10(x) for x in points], profile), f)
            f.close()

    if which=='all':
        opt = Optimizer(True, True, quant_data=Ybar_data, quant_func=Ybar_function, qual_data=A_data, qual_func=A_func,
                        qual_0=activity0, sensitivity=1.5, qual_c=qual_c)
        profile = opt.profile_likelihood(const_param, points)
        plt.plot([np.log10(x) for x in points], profile, '^-',label='Combined', color=cbm[2])
        if savestem:
            f = open(savestem+'_all', 'wb')
            pickle.dump(([np.log10(x) for x in points], profile), f)
            f.close()

    if which=='alltot':
        activity0_sols = sp.solve([2 * RR + R - 50, RR / R ** 2 - oracle_vals[K1]], [R, RR])
        activity0 = activity0_sols[1][1]
        A_data = gen_expt_data_set('ItotA', n=qualpts)
        A_func = ode_Itot_A_func
        opt = Optimizer(False, True, qual_data=A_data, qual_func=A_func, qual_0=activity0, sensitivity=1,
                        qual_c=qual_c)
        profile = opt.profile_likelihood(const_param, points)
        plt.plot([np.log10(x) for x in points], profile, 'o-', label='Qual. Itot vs A')
        if savestem:
            f = open(savestem + '_alltot', 'wb')
            pickle.dump(([np.log10(x) for x in points], profile), f)
            f.close()


    # plt.xlabel('Log('+const_param+')')
    plt.xlabel("$\\mathregular{log_{10}(K_%s)}$" % const_param[1])
    plt.ylabel('Objective')
    plt.tight_layout()

    if which!='qual':
        plt.ylim((-0.1, 1))

    plt.axvline(x=np.log10(ground_const[eval(const_param)]), color=cbm[3])
    plt.legend(fontsize=12)


def plot_saved_profiles(const_param, stem, cutoff=0.15):

    def cutoff_calc(x,y):
        inds = np.argwhere(y < cutoff).flatten()
        return [x[inds[0]], x[inds[-1]]]

    plt.figure(figsize=(4, 3))

    f = open(stem+'quant', 'rb')
    x, y = pickle.load(f)
    y = y - np.min(y)
    plt.plot(x, y, '-', label='Quantitative', color=cbm[0], linewidth=4, dashes=[4, 2])
    f.close()
    print("Quant bounds: %s" % cutoff_calc(x, y))

    f = open(stem + 'qual', 'rb')
    x, y = pickle.load(f)
    y = y - np.min(y)
    plt.plot(x, y, '-', label='Qualitative', color=cbm[1], linewidth=4, dashes=[4, 1, 1, 1])
    f.close()
    print("Qual bounds: %s" % cutoff_calc(x, y))

    f = open(stem + 'all', 'rb')
    x, y = pickle.load(f)
    y = y - np.min(y)
    plt.plot(x, y, '-', label='Combined', color=cbm[2], linewidth=4)
    f.close()
    print("Tot bounds: %s" % cutoff_calc(x, y))
    # Incantation to correct the spacing of ticks on x axis
    ax = plt.gca()
    start, end = ax.get_xlim()
    ax.xaxis.set_ticks(np.arange(np.ceil(start), end, 1.))
    plt.axvline(x=np.log10(ground_const[eval(const_param)]), linewidth=3, color=cbm[3])
    plt.ylim((-0.05, 0.5))
    # plt.legend(fontsize=12, handlelength=5)
    plt.xlabel("$\\mathregular{log_{10}(K_%s)}$" % const_param[1])
    plt.ylabel('Objective')
    plt.tight_layout()
