import pset
import numpy as np
import pickle


class ParallelDifferentialEvolution():
    """
    Implements the parallelized, island-based differential evolution algorithm
    described in Penas et al 2015.

    In some cases, I had to make my own decisions for specifics I couldn't find in the original paper. Namely:
    At each migration, a user-defined number of individuals are migrated from each island. For each individual, a
    random index is chosen; the same index for all islands. A random permutation is used to redistribute individuals
    with that index to different islands.

    Each island performs its migration individually, on the first callback when all islands are ready for that
    migration. It receives individuals from the migration iteration, regardless of what the current iteration is.
    This can sometimes lead to wasted effort.
    For example, suppose migration is set to occur at iteration 40, but island 1 has reached iteration 42 by the time
    all islands reach 40. Individual j on island 1 after iteration 42 gets replaced with individual j on island X
    after iteration 40. Some other island Y receives individual j on island 1 after iteration 40.

    """

    def __init__(self, variables, num_per_island, num_islands, migrate_every, mutation_rate, mutation_factor,
                 max_iterations, num_to_migrate, save_every=100):
        """

        In the future, this function should take a config object instead of all
        these parameters.

        Required keys in config:
        population_size
        num_islands*
        max_generations
        mutation_rate
        mutation_factor
        migrate_every
        num_to_migrate

        and all your free parameters

        :param expdata: Experimental data to be fit
        :param objective: Objective function
        :param num_per_island: Number of individuals per island
        :param num_islands: The number of islands over which to distribute the individuals,
        corresponding to the number of parallel processes possible.
        :param migrate_every: Migrate individuals between islands every x generations.
        """
        self.num_per_island = num_per_island
        self.num_islands = num_islands
        self.migrate_every = migrate_every
        self.mutation_rate = mutation_rate
        self.mutation_factor = mutation_factor
        self.variables = variables  # tuples (name, min, max)
        self.bounds = {v[0]: (v[1], v[2]) for v in variables}  # {var_name:(min, max)}
        self.max_iterations = max_iterations
        self.num_to_migrate = num_to_migrate
        self.save_every = save_every

        self.island_map = dict()  # Maps each proposed PSet to its location (island, individual_i)
        self.iter_num = [0] * num_islands  # Count the number of completed iterations on each island
        self.waiting_count = []  # Count of the number of PSets that are pending evaluation on the current iteration of each island.
        self.individuals = []  # Nested list; individuals[i][j] gives individual j on island i.
        self.proposed_individuals = []  # Nested list of the same shape, gives individuals proposed for replacement in next generation
        self.fitnesses = []  # Nested list of same shape, gives fitness of each individual
        self.migration_ready = [0] * num_islands  # What migration number is each island ready for
        self.migration_done = [0] * num_islands  # What migration number has each island completed


        # These variables store data related to individual migrations.
        # Each one has migration number as keys. When the first island starts migration, the required entries are
        # created. When the last island completes migration, they are deleted to keep these structures small.
        self.migration_transit = dict()  # Store (PSet, fitness) tuples here that are getting migrated - one list per island
        self.migration_indices = dict()  # Which individual numbers are migrating in migration i - a single tuple for
        # each migration, used for all islands
        self.migration_perms = dict()  # How do we rearrange between islands on migration i?
        # For each migration, a list of num_to_migrate permutations of range(num_islands)


        self.strategy = 'rand1'  # Customizable later

    def start_run(self):

        # Initialize random individuals
        # self.proposed_individuals = [[pset.PSet({v[0]: np.random.uniform(v[1], v[2]) for v in self.variables}, True)
        #                               for i in range(self.num_per_island)]
        #                              for j in range(self.num_islands)]

        # Initialize the individuals on a latin hypercube
        rands = latin_hypercube(self.num_per_island * self.num_islands, len(self.variables))
        # rands = latin_hypercube(30, len(self.variables)) # For testing
        psets = []
        for row in rands:
            # Convert the 0 to 1 random numbers to the specified variable range variables[i][1] to variables[i][2]
            d = {self.variables[i][0]: self.variables[i][1] + row[i] * (self.variables[i][2] - self.variables[i][1])
                 for i in range(len(row))}
            psets.append(pset.PSet(d, allow_negative=True))

        self.proposed_individuals = [psets[i*self.num_per_island: (i+1)*self.num_per_island]
                                     for i in range(self.num_islands)]

        # Initialize the individual list to empty, will be filled with the proposed_individuals once their fitnesses
        # are computed.
        self.individuals = [[None
                             for i in range(self.num_per_island)]
                            for j in range(self.num_islands)]

        # Set all fitnesses to Inf, guaraneeting a replacement by the first proposed individual
        self.fitnesses = [[np.Inf
                           for i in range(self.num_per_island)]
                          for j in range(self.num_islands)]

        for i in range(len(self.proposed_individuals)):
            for j in range(len(self.proposed_individuals[i])):
                self.island_map[self.proposed_individuals[i][j]] = (i, j)

        self.waiting_count = [self.num_per_island] * self.num_islands

        return [ind for island in self.proposed_individuals for ind in island]

    def got_result(self, pset, score):
        """
        Called when a simulation run finishes

        This is not thread safe - the Scheduler must ensure only one process at a time enters
        this function.
        (or, I should rewrite this function to make it thread safe)

        :param pset:
        :param simdata:
        :return:
        """

        # Calculate the fitness of this individual, and replace if it is better than the previous one.
        island, j = self.island_map.pop(pset)
        fitness = score
        if fitness < self.fitnesses[island][j]:
            self.individuals[island][j] = pset
            self.fitnesses[island][j] = fitness
        self.waiting_count[island] -= 1

        # Determine if the current iteration is over for the current island
        if self.waiting_count[island] == 0:

            self.iter_num[island] += 1
            if self.iter_num[island] == self.max_iterations:
                # Submit no more jobs for this island
                # Once all islands reach this, simulation is over.
                return []

            if self.iter_num[island] % self.migrate_every == 0:
                # This island prepares for migration
                migration_num = int(self.iter_num[island] / self.migrate_every)
                if max(self.migration_ready) < migration_num:
                    # This is the first island to reach this migration.
                    # Need to set global parameters for this migration.
                    self.migration_transit[migration_num] = [list() for i in range(self.num_islands)]
                    self.migration_indices[migration_num] = np.random.choice(range(self.num_per_island),
                                                                             size=self.num_to_migrate, replace=False)
                    self.migration_perms[migration_num] = [np.random.permutation(self.num_islands)
                                                           for i in range(self.num_to_migrate)]
                    print('Island %i just set up the migration with the following specs')
                    print(self.migration_transit)
                    print(self.migration_indices)
                    print(self.migration_perms)

                # Send the required PSets to migration_transit
                for j in self.migration_indices[migration_num]:
                    self.migration_transit[migration_num][island].append((self.individuals[island][j],
                                                                          self.fitnesses[island][j]))
                # Tell other islands that this one is ready for this migration.
                self.migration_ready[island] = migration_num

            if self.migration_done[island] < min(self.migration_ready):
                # This island performs a migration
                print('Island %i is migrating! The migration specs are as follows' % island)
                print(self.migration_transit)
                print(self.migration_indices)
                print(self.migration_perms)
                migration_num = self.migration_done[island] + 1

                # Fetch the appropriate new individuals from migration_transit
                for migrater_index in range(self.num_to_migrate):
                    j = self.migration_indices[migration_num][migrater_index]  # Index of the individual
                    newisland = self.migration_perms[migration_num][migrater_index][island]
                    self.individuals[island][j], self.fitnesses[island][j] = \
                        self.migration_transit[migration_num][newisland][migrater_index]

                    print('Island %i gained new individual with fitness %f' % (island, self.fitnesses[island][j]))

                self.migration_done[island] = migration_num
                if min(self.migration_done) == migration_num:
                    # This is the last island to complete this migration
                    # Delete the migration data to free space.
                    del self.migration_transit[migration_num]
                    del self.migration_perms[migration_num]
                    del self.migration_indices[migration_num]

            # Set up the next generation
            for jj in range(self.num_per_island):
                self.proposed_individuals[island][jj] = self.new_individual(island)
                self.island_map[self.proposed_individuals[island][jj]] = (island, jj)

            self.waiting_count[island] = self.num_per_island

            if self.iter_num[island] % 20 == 0:
                print('Completed %i iterations' % self.iter_num[island])
                print(sorted(self.fitnesses[island]))
            if self.iter_num[island] % self.save_every == 0:
                f = open('results/diffev_%i_%i' % (island, self.iter_num[island]), 'wb')
                pickle.dump((self.individuals, self.fitnesses), f)
                f.close()

            # Convergence check
            if np.max(self.fitnesses) / np.min(self.fitnesses) < 1.002:
                return 'STOP'

            return self.proposed_individuals[island]

        else:
            # Add no new jobs, wait for this generation to complete.
            return []

    def new_individual(self, island):
        """
        Create a new individual for the specified island, according to the set strategy

        :param island:
        :return:
        """

        # Choose a starting parameter set (either the best one, or a random one)
        if self.strategy in ['rand1']:
            base = np.random.choice(self.individuals[island])
        else:
            raise NotImplementedError('Please select one of the strategies from our extensive list of options: rand1')

        # Choose other parameter sets to cross over
        others = np.random.choice(self.individuals[island], 2, replace=False)

        # Iterate through parameters; decide whether to mutate or leave the same.
        new_pset_dict = dict()
        for p in base.keys():
            if np.random.random() < self.mutation_rate:
                new_pset_dict[p] = (max(self.bounds[p][0], min(self.bounds[p][1],
                                    base[p] + self.mutation_factor * (others[0][p] - others[1][p]))))
            else:
                new_pset_dict[p] = base[p]

        return pset.PSet(new_pset_dict, allow_negative=True)


class ScatterSearch:
    """
    Implements ScatterSearch as described in the introduction of Penas et al 2017 (but not the fancy parallelized
    version from that paper).
    Uses the individual combination method described in Egea et al 2009

    """

    def __init__(self, variables, popsize, maxiters, saveevery):

        self.variables = variables
        self.popsize = popsize
        self.maxiters = maxiters
        self.saveevery = saveevery

        self.pending = dict() # {pendingPSet: parentPSet}
        self.received = dict() # {parentPSet: [(donependingPSet, score)]
        self.refs = [] # (refPset, score)
        self.stuckcounter = dict()
        self.iteration = 0
        self.local_mins = [] # (Pset, score) pairs that were stuck for 5 gens, and so replaced.
        self.reserve = []


    def start_run(self):
        # Generate big number = 10 * variable_count initial individuals.
        rands = latin_hypercube(10*len(self.variables), len(self.variables))
        # rands = latin_hypercube(30, len(self.variables)) # For testing
        psets = []
        for row in rands:
            # Convert the 0 to 1 random numbers to the specified variable range variables[i][1] to variables[i][2]
            d = {self.variables[i][0]: self.variables[i][1] + row[i]*(self.variables[i][2] - self.variables[i][1])
                 for i in range(len(row))}
            psets.append(pset.PSet(d, allow_negative=True))

        # Generate a latin hypercube distributed "reserve". When we need a random new individual, pop one from here
        # so we aren't repeating ground.
        rands2 = latin_hypercube(self.maxiters, len(self.variables))
        for row in rands2:
            # Convert the 0 to 1 random numbers to the specified variable range variables[i][1] to variables[i][2]
            d = {self.variables[i][0]: self.variables[i][1] + row[i]*(self.variables[i][2] - self.variables[i][1])
                 for i in range(len(row))}
            self.reserve.append(pset.PSet(d, allow_negative=True))

        self.pending = {p: None for p in psets}
        self.received = {None: []}
        return psets

    def round_1_init(self):
        start_psets = sorted(self.received[None], key=lambda x: x[1])
        # Half is the top of the list, half is random.
        topcount = int(np.ceil(self.popsize / 2.))
        randcount = int(np.floor(self.popsize / 2.))
        self.refs = start_psets[:topcount]
        randindices = np.random.choice(range(topcount, len(start_psets)), randcount, replace=False)
        for i in randindices:
            self.refs.append(start_psets[i])
        self.stuckcounter = {r[0]: 0 for r in self.refs}

    def got_result(self, ps, score):
        """
        Called when a simulation run finishes

        This is not thread safe - the Scheduler must ensure only one process at a time enters
        this function.
        (or, I should rewrite this function to make it thread safe)

        :param ps:
        :param simdata:
        :return:
        """
        parent = self.pending[ps]
        self.received[parent].append((ps, score))
        del self.pending[ps]

        if len(self.pending) == 0:
            # All of this generation done, make the next list of psets

            if None in self.received:
                # This is the initialization round, special case
                self.round_1_init()
            else:
                # 1) Replace parent with highest scoring child
                for i in range(len(self.refs)):
                    best_child = min(self.received[self.refs[i][0]], key=lambda x: x[1])
                    if best_child[1] < self.refs[i][1]:
                        del self.stuckcounter[self.refs[i][0]]
                        self.stuckcounter[best_child[0]] = 0
                        self.refs[i] = best_child
                    else:
                        self.stuckcounter[self.refs[i][0]] += 1
                        if self.stuckcounter[self.refs[i][0]] >= 5:
                            del self.stuckcounter[self.refs[i][0]]
                            self.local_mins.append(self.refs[i])
                            # For output. Not the most efficient, but not in a performance-critical section
                            self.local_mins = sorted(self.local_mins, key=lambda x: x[1])
                            self.local_mins = self.local_mins[:self.popsize] # So this doesn't get huge

                            # Pick a new random pset
                            if len(self.reserve) > 0:
                                new_pset = self.reserve.pop()
                            else:
                                new_pset = pset.PSet({v[0]: np.random.uniform(v[1], v[2]) for v in self.variables}, allow_negative=True)
                            self.refs[i] = (new_pset, np.inf)  # For simplicity, assume its score is awful
                            self.stuckcounter[new_pset] = 0


            # 2) Sort the refs list by quality.
            self.refs = sorted(self.refs, key=lambda x: x[1])
            print(self.iteration)
            print('Current: ' + str([x[1] for x in self.refs]))
            print('Archive: ' + str([x[1] for x in self.local_mins]))

            if self.iteration % self.saveevery == 0:
                f = open('results/scatter_%i' % self.iteration, 'wb')
                pickle.dump((self.refs, self.local_mins), f)
                f.close()

            self.iteration += 1
            if self.iteration == self.maxiters:
                return 'STOP'

            # 3) Do the combination antics to generate new candidates
            query_psets = []
            for pi in range(self.popsize): # parent index
                for hi in range(self.popsize): # helper index
                    if pi == hi:
                        continue
                    newdict = dict()
                    for var in self.variables:
                        v = var[0]
                        d = (self.refs[hi][0][v] - self.refs[pi][0][v]) / 2.
                        alpha = np.sign(hi-pi)
                        beta = (abs(hi-pi) - 1) / (self.popsize - 2)
                        c1 = self.refs[pi][0][v] - d*(1 + alpha*beta)
                        c2 = self.refs[pi][0][v] + d*(1 - alpha*beta)
                        newval = np.random.uniform(c1, c2)
                        newdict[v] = max(min(newval, var[2]), var[1])
                    newpset = pset.PSet(newdict, allow_negative=True)
                    query_psets.append(newpset)
                    self.pending[newpset] = self.refs[pi][0]
            self.received = {r[0]: [] for r in self.refs}
            return query_psets

        else:
            return []



def latin_hypercube(nsamples, ndims):
    """
    Latin hypercube sampling.
    This code was dug out of the scipy.optimize.differentialevolution source code, and converted to work in the
    general case (which surprisingly, does not exist within scipy)

    Initializes the population with Latin Hypercube Sampling.
    Latin Hypercube Sampling ensures that each parameter is uniformly
    sampled over its range.

    Returns a nsamples by ndims array, with entries in the range [0,1]
    You'll have to rescale them to your actual param ranges.
    """
    rng = np.random

    # Each parameter range needs to be sampled uniformly. The scaled
    # parameter range ([0, 1)) needs to be split into
    # `self.num_population_members` segments, each of which has the following
    # size:
    segsize = 1.0 / nsamples

    # Within each segment we sample from a uniform random distribution.
    # We need to do this sampling for each parameter.
    samples = (segsize * rng.random_sample((nsamples, ndims))

    # Offset each segment to cover the entire parameter range [0, 1)
               + np.linspace(0., 1., nsamples,
                             endpoint=False)[:, np.newaxis])

    # Create an array for population of candidate solutions.
    population = np.zeros_like(samples)

    # Initialize population of candidate solutions by permutation of the
    # random samples.
    for j in range(ndims):
        order = rng.permutation(range(nsamples))
        population[:, j] = samples[order, j]

    return population
