import numpy as np
import re
from numba import jit


class ConstraintParser:
    """
    Parses a constraint file in the new Constraint Language

    """

    @staticmethod
    def parse(filename):
        """
        Parse the specified .con file written in Constraint Language

        Returns a dictionary object with keys giving tuples of Cases and values giving a list of Constraints

        :param filename:
        :return:
        """

        f = open(filename)
        current_case = tuple()
        constraints = dict()

        for line in f:
            if line[0] == '#':
                continue  # Comment
            tokens = re.split('\s+', line)
            if tokens == ['', '']:
                continue  # Blank line
            if tokens[0] == 'with':
                # New case set
                # each token except the first 'with' and the last '' should represent one case.
                current_case = tuple([Case(s) for s in tokens[1:-1]])
                continue
            elif tokens[0] != '':
                # Line is not indented, so no longer falls under the previous case
                current_case = tuple()
            else:
                tokens.pop(0)

            # Parse the constraint on this line.
            or_list = []  # For OrConstraints of the form constraint1 or constraint2
            inner_or_list = []  # For AlwaysOr or NotAlwaysOr constraints.
            inner_or_weights = []
            while True:  # Iterate potentially over multiple conditions if this line has or's in it.
                ineq = tokens.pop(0)
                keyword = tokens.pop(0)
                if keyword == 'when':
                    if len(inner_or_list) > 0:
                        raise ParseError('Unsupported syntax: <cond1> or <cond2> when')
                    when_condition = tokens.pop(0)
                    c_new = WhenConstraint(ineq, when_condition)
                elif keyword == 'until':
                    if len(inner_or_list) > 0:
                        raise ParseError('Unsupported syntax: <cond1> or <cond2> until')
                    until_condition = tokens.pop(0)
                    c_new = UntilConstraint(ineq, until_condition)
                elif keyword == 'always':
                    if len(inner_or_list) > 0:
                        c_new = AlwaysOrConstraint(inner_or_list, inner_or_weights)
                    else:
                        c_new = AlwaysConstraint(ineq)
                elif keyword == 'notalways':
                    if len(inner_or_list) > 0:
                        raise ParseError('Unsupported syntax: <cond1> or <cond2> notalways')
                    else:
                        c_new = NotAlwaysConstraint(ineq)
                elif keyword == 'or':
                    inner_or_list.append(ineq)
                    inner_or_weights.append(1.0)
                    continue
                elif keyword == 'weight':
                    # Should only get here with the AlwaysOr syntax.
                    w = tokens.pop(0)
                    inner_or_list.append(ineq)
                    inner_or_weights.append(float(w))
                    next = tokens.pop(0)
                    if next == 'or':
                        continue
                    elif len(inner_or_list) == 1:
                        ParseError('Keyword "weight" immediately after inequality. This syntax is not valid unless you '
                                   'are defining an AlwaysOr constraint.')
                    elif next == 'always':
                        c_new = AlwaysOrConstraint(inner_or_list, inner_or_weights)
                    else:
                        ParseError(
                            'Invalid keyword "' + next + '" after "<inequality> weight <x>". That syntax is only '
                                                         'used for an AlwaysOr constraint, so valid keywords here are "always" or "or"')

                else:
                    raise ParseError('Expected a constraint keyword instead of "' + keyword + '"')

                next = tokens.pop(0)

                if next == 'penalty':
                    if keyword != 'when':
                        raise ParseError("Only 'when' constraints should receive the 'penalty' keyword")
                    # Replace the new WhenConstraint constraint with an AltWhenConstraint
                    c_new = AltWhenConstraint(ineq, when_condition, custom_penalty=tokens.pop(0))
                    next = tokens.pop(0)

                if next == 'weight':
                    c_new.weight = float(tokens.pop(0))
                    next = tokens.pop(0)

                if next == 'min':
                    if type(c_new) != AltWhenConstraint:
                        raise ParseError("Keyword 'min' should only be used after 'penalty' in a 'when' constraint")
                    c_new.min_penalty = float(tokens.pop(0))
                    next = tokens.pop(0)

                if next == 'never':
                    if keyword != 'when':
                        raise ParseError("Only 'when' constraints should receive the 'never' keyword")
                    c_new.never_penalty = float(tokens.pop(0))
                    next = tokens.pop(0)

                or_list.append(c_new)
                inner_or_list = []
                inner_or_weights = []

                if next == '':
                    break
                elif next == 'or':
                    pass
                else:
                    raise ParseError('Unexpected token ' + next)

            # Create the OrConstraint if applicable
            if len(or_list) > 1:
                c_new = OrConstraint(or_list)

            if current_case in constraints:
                constraints[current_case].append(c_new)
            else:
                constraints[current_case] = [c_new]
        return constraints

    @staticmethod
    def parse_skeleton(filename):
        """
        Parse a skeleton constraint file.
        This is a file that does not contain constraints; instead just has a string associated with each case

        Returns a dictionary object with keys giving tuples of Cases and values giving strings
        :param filename:
        :return:
        """
        f = open(filename)
        current_case = tuple()
        constraints = dict()

        for line in f:
            if line[0] == '#':
                continue  # Comment
            tokens = re.split('\s+', line)
            if len(current_case) > 0:
                # Tabulate the string for the case that we just read
                constraints[current_case] = tokens[0]
                current_case = tuple()
                continue
            if tokens == ['', '']:
                continue  # Blank line
            if tokens[0] == 'with':
                # New case set
                # each token except the first 'with' and the last '' should represent one case.
                current_case = tuple([Case(s) for s in tokens[1:-1]])
                continue
            else:
                raise ParseError('Unrecognized line %s' % line)

        return constraints

    @staticmethod
    def tokenize(line):
        """
        Splits a line of Constraint Language into its relevant pieces

        :param line:
        :return: list of tokens (strings)
        """
        pass


class Case:
    """
    Class representing a Case in which some parameters are altered in order to consider the constraint

    After initialization, contains one fields for external use
    Case.funcstr: String meant to be called with exec() when setting up the case.

    """

    def __init__(self, casestr):
        """
        Creates a new Case with the given string

        The input string takes the form of a python command to be executed to create the case
        Example 'param1=4*param1+2*param2'
        Supports only arithmetic operations, as all words will be mistaken for variable names.

        Use the canonical variable names as they appear in the C++ file for params, or in INITIAL_CONDS_NAMES for
        initial conditions

        The constructor will automatically convert these names to keys for the params dict, which is what the
        Optimizer needs to run them

        :param casestr: a string describing the case
        :type casestr: str
        """

        self.funcstr = re.sub('[A-Za-z_][A-Za-z0-9_]*', self.subfunc, casestr)
        self.casestr = casestr

    @staticmethod
    def subfunc(match):
        """ Helper function for the re matching."""
        return "params['" + match.group() + "']"

    def __str__(self):
        return self.casestr

    def __repr__(self):
        return self.casestr

    def __hash__(self):
        return hash(self.funcstr)

    def __eq__(self, other):
        return self.casestr == other.casestr


class Constraint:
    """
    Abstract class representing an optimization constraint with a penalty for violating the constraint
    """

    def __init__(self):
        pass

    def penalty(self, data):
        """
        penalty function for violating the constraint. Returns 0 if constraint is satisfied, or a positive value
        if the constraint is volated. Implementation depends on the type of constraint.

        :param data: TysonData object
        :type data: TysonData
        :return: float value of penalty
        """

        raise NotImplementedError('Subclasses of Constraint must override penalty()')

    def _parse_inequality(self, ineq):
        """
        Set this constraint's inequality according to the specified inequality string
        Updates the target_var, target_val, and sign fields.

        :param ineq: String representing the inequality
        :return:
        """
        sign = None
        for asign in ['<=', '>=', '<', '>']:
            if asign in ineq:
                sign = asign
                break
        if sign is None:
            raise ParseError('Unable to parse inequality ' + ineq)
        parts = ineq.split(sign)
        if len(parts) != 2:
            raise ParseError('Unable to parse inequality ' + ineq)
        self.target_var = parts[0]
        try:
            self.target_val = float(parts[1])
        except ValueError:
            # Alternate option to give a variable name here.
            self.target_val = parts[1]
        self.sign = sign

    def str_info(self, penalty=None):
        return (' ({:.3g}) '.format(penalty)) + self.target_var + self.sign + float_clean(self.target_val)

    def str_info_latex(self, penalty=None):
        return latex_escape(('({:.3g}) '.format(penalty)) + '$' + self.target_var + self.sign + float_clean(self.target_val) + '$')


class WhenConstraint(Constraint):
    def __init__(self, inequality, when_condition, never_penalty=1e5):
        """
        Creates a new constraint of the form

        X1>value when X2=value

        :param inequality: String specifying the inequality, eg 'X1>4'
        :param when_condition: String specifying the when condition eg 'X2=3'
        :param never_penalty: The large penalty for never satisfying the when condition within the data
        """

        super().__init__()
        self.never_penalty = never_penalty
        self._parse_inequality(inequality)

        parts = when_condition.split('=')
        if len(parts) != 2:
            raise ParseError("Unable to parse 'when' condition " + when_condition)
        self.when_var = parts[0]
        self.when_val = float(parts[1])
        self.weight = 1.0

    def penalty(self, data):
        """
        Compute the when penalty.
        If the when condition is never met, the penalty is the large never_penalty times the closest approach to the
        when condition.

        In the normal case, the penalty is 0 if the constraint is met, or the distance from the constraint if it
        is failed, at the index where the when condition is first met.

        Edge cases where the the when condition is hit exactly at a particular index are not currently implemented
        correctly

        Now supports declaring a weight of the constraint by setting the .weight attribute. This is a constant
        multiplier on the penalty function

        :param data:
        :type data: TysonData
        :return: float penalty falue
        """

        # Find the index where the when condition is met
        when_col = data[self.when_var] < self.when_val
        first_true = find_first(True, when_col)
        first_false = find_first(False, when_col)

        if first_true == -1:
            # When_var is never lower than when_val
            return self.never_penalty * np.min(data[self.when_var] - self.when_val)

        if first_false == -1:
            # When_var is never higher than when_val
            return self.never_penalty * np.min(self.when_val - data[self.when_var])

        keyindex = max(first_false, first_true)

        # Check the appropriate index
        diff = data[self.target_var][keyindex] - self.target_val  # positive if data > target

        if self.sign == '>' or self.sign == '>=':
            diff = -diff

        return max(0., diff) * self.weight

    def str_info(self, penalty=None):
        return super().str_info(penalty) + ' when ' + self.when_var + '=' + str(self.when_val)

    def str_info_latex(self, penalty=None):
        return super().str_info_latex(penalty) + latex_escape(' when $' + self.when_var + '=' + str(self.when_val) + '$')

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ $%s(\\tau) %s %s$'.format(penalty) %
                            (self.target_var, self.sign, float_clean(self.target_val)))


class AltWhenConstraint(Constraint):
    def __init__(self, inequality, when_condition, custom_penalty, never_penalty=1e5, min_penalty=0.):
        """
        Creates a new constraint of the form

        X1>value when X2=value penalty X3<value min 1

        Here, we switch the normal penalty (from looking at X1) to a custom penalty by looking at X3.
        This is useful if X1 is a switch-type variable controlled by X3, making X3 a better non-discrete metric of how
        much you missed the constraint.
        Note that if the original constraint is violated but the penalty is not, the penalty will be min_penalty

        :param inequality: String specifying the inequality, eg 'X1>4'
        :param when_condition: String specifying the when condition eg 'X2=3'
        :param custom_penalty: String specifying another inequality which is used to compute the penalty, eg X3<2
        :param never_penalty: The large penalty for never satisfying the when condition within the data
        :param min_penalty: A constant penalty for violating the constraint, regardless of whether custom_penalty is
        violated. Useful for avoiding shenanigans in which the constraint is dodged by having the penalty be close to
        0 when the main inequality is violated.
        """

        super().__init__()
        self.never_penalty = never_penalty
        # The parse function auto-saves values into certain fields. We call that, then move the values to the fields
        # where we actually want them.
        self._parse_inequality(custom_penalty)
        self.custom_var = self.target_var
        self.custom_val = self.target_val
        self.custom_sign = self.sign
        # Now call _parse_inequality on the actual inequality, overwriting those 3 values to the actual values.
        self._parse_inequality(inequality)

        parts = when_condition.split('=')
        if len(parts) != 2:
            raise ParseError("Unable to parse 'when' condition " + when_condition)
        self.when_var = parts[0]
        self.when_val = float(parts[1])
        self.min_penalty = min_penalty
        self.weight = 1.0

    def penalty(self, data):
        """
        Compute the when penalty.
        If the when condition is never met, the penalty is the large never_penalty times the closest approach to the
        when condition.

        In the normal case, the penalty is 0 if the constraint is met, or the distance from the constraint if it
        is failed, at the index where the when condition is first met.

        Edge cases where the the when condition is hit exactly at a particular index are not currently implemented
        correctly

        Now supports declaring a weight of the constraint by setting the .weight attribute. This is a constant
        multiplier on the penalty function

        :param data:
        :type data: TysonData
        :return: float penalty falue
        """

        # Find the index where the when condition is met
        when_col = data[self.when_var] < self.when_val
        first_true = find_first(True, when_col)
        first_false = find_first(False, when_col)

        if first_true == -1:
            # When_var is never lower than when_val
            return self.never_penalty * np.min(data[self.when_var] - self.when_val)

        if first_false == -1:
            # When_var is never higher than when_val
            return self.never_penalty * np.min(self.when_val - data[self.when_var])

        # Todo: Currently contains a problem-specific hack.
        # We subtract 1 from keyindex because, for the tests of DIV_COUNT I'll be using this feature for, the other
        # flags reset on the same time tick that DIV_COUNT goes up; we actually care about whether they were on at
        # the tick before that.
        keyindex = max(first_false, first_true) - 1

        # Todo: Currently only this constraint type properly handles the case of self.target_val actually being a variable name.
        if type(self.target_val) == str:
            theval = data[self.target_val][keyindex]
        else:
            theval = self.target_val

        # Check the appropriate index
        diff = data[self.target_var][keyindex] - theval  # positive if data > target

        if self.sign == '>' or self.sign == '>=':
            diff = -diff

        if diff <= 0:
            return 0.  # Main inequality was not violated.

        if type(self.custom_val) == str:
            theval = data[self.custom_val][keyindex]
        else:
            theval = self.custom_val

        # Check the custom penalty at that index
        diff = data[self.custom_var][keyindex] - theval  # positive if data > target
        if self.custom_sign == '>' or self.custom_sign == '>=':
            diff = -diff

        return max(0., diff) * self.weight + self.min_penalty

    def str_info(self, penalty=None):
        return super().str_info(penalty) + ' when ' + self.when_var + '=' + str(self.when_val)

    def str_info_latex(self, penalty=None):
        return super().str_info_latex(penalty) + latex_escape(' when $' + self.when_var + '=' + str(self.when_val) + '$')

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ $%s(\\tau) %s %s$'.format(penalty) %
                            (self.target_var, self.sign, float_clean(self.target_val)))


class UntilConstraint(Constraint):
    def __init__(self, inequality, when_condition):
        """
        Creates a new constraint of the form

        X1>value until X2=value

        :param inequality: String specifying the inequality, eg 'X1>4'
        :param when_condition: String specifying the when condition eg 'X2=3'
        """

        super().__init__()
        self._parse_inequality(inequality)

        parts = when_condition.split('=')
        if len(parts) != 2:
            raise ParseError("Unable to parse 'until' condition " + when_condition)
        self.when_var = parts[0]
        self.when_val = float(parts[1])
        self.weight = 1.0

    def penalty(self, data):
        """
        Compute the until penalty.

        The until constraint indicates that the constraint must be satisfied at every time point until the condition
        is satisfied.
        If the until condition is never met, there's no penalty for that, but then the constraint must be satisfied
        for the entire simulation.

        The penalty is given by the worst violation of the constraint within the time range that the constraint is active.

        Edge cases where the the when condition is hit exactly at a particular index are not currently implemented
        correctly

        Now supports declaring a weight of the constraint by setting the .weight attribute. This is a constant
        multiplier on the penalty function

        :param data:
        :type data: TysonData
        :return: float penalty falue
        """

        # Find the index where the when condition is met
        when_col = data[self.when_var] < self.when_val
        first_true = find_first(True, when_col)
        first_false = find_first(False, when_col)

        if first_true == -1 or first_false == -1:
            keyindex = len(data[self.when_var])
        else:
            keyindex = max(first_false, first_true)

        # Check the appropriate indices (everything up to keyindex). Like an AlwaysConstraint for just this range.
        diffs = data[self.target_var][:keyindex] - self.target_val  # positive if data > target

        if self.sign == '>' or self.sign == '>=':
            diffs = -diffs

        worstdiff = np.max(diffs)

        return max(0., worstdiff) * self.weight

    def str_info(self, penalty=None):
        return super().str_info(penalty) + ' until ' + self.when_var + '=' + str(self.when_val)

    def str_info_latex(self, penalty=None):
        return super().str_info_latex(penalty) + latex_escape(' until $' + self.when_var + '=' + str(self.when_val) + '$')

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ $\\max(%s(t)) %s %s$'.format(penalty) %
                            (self.target_var, sign_format(self.sign), float_clean(self.target_val)))


class AlwaysConstraint(Constraint):
    def __init__(self, inequality):
        super().__init__()
        self._parse_inequality(inequality)
        self.weight = 1.0

    def penalty(self, data):
        """
        Compute the always penalty
        The penalty is given by the worst miss of the constraint over the entire data column


        :param data: TysonData object
        :return:
        """

        diffs = data[self.target_var] - self.target_val
        if self.sign == '>' or self.sign == '>=':
            diffs = -diffs

        worstdiff = np.max(diffs)

        return max(0., worstdiff) * self.weight

    def str_info(self, penalty=None):
        return super().str_info(penalty) + ' always'

    def str_info_latex(self, penalty=None):
        return super().str_info_latex(penalty) + ' always'

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ $\\max(%s(t)) %s %s$'.format(penalty) %
                            (self.target_var, sign_format(self.sign), float_clean(self.target_val)))


class NotAlwaysConstraint(Constraint):
    def __init__(self, inequality):
        super().__init__()
        self._parse_inequality(inequality)
        self.weight = 1.0

    def penalty(self, data):
        """
        Compute the notalways penalty
        The penalty is given by the smallest miss of the opposite of the constraint, over the entire data column

        :param data: TysonData object
        :return: float
        """

        diffs = data[self.target_var] - self.target_val

        if self.sign == '<' or self.sign == '<=':
            diffs = -diffs

        bestdiff = np.min(diffs)

        return max(0., bestdiff) * self.weight

    def str_info(self, penalty=None):
        return super().str_info(penalty) + ' notalways'

    def str_info_latex(self, penalty=None):
        return super().str_info_latex(penalty) + ' notalways'

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ $\\max(%s(t)) %s %s$'.format(penalty) %
                            (self.target_var, sign_format(self.sign, flip=True), float_clean(self.target_val)))

class OrConstraint(Constraint):
    def __init__(self, constraintlist):
        super().__init__()
        self.constraintlist = constraintlist

    def penalty(self, data):
        """
        Compute the or penalty, as the minimum penalty value among all the available options

        :param data:
        :return:
        """
        return min([c.penalty(data) for c in self.constraintlist])


class AlwaysOrConstraint(Constraint):
    """
    Constraint which says that at all times, at least one of the following must be true,
    but which one is true can vary over time.

    """

    def __init__(self, constraintlist, weightlist):
        super().__init__()
        self.constraintlist = constraintlist
        self.weightlist = weightlist
        self.target_var_list = []
        self.target_val_list = []
        self.sign_list = []
        for ineq in constraintlist:
            self._parse_inequality(ineq)
            self.target_var_list.append(self.target_var)
            self.target_val_list.append(self.target_val)
            self.sign_list.append(self.sign)

    def penalty(self, data):
        """
        Compute the always penalty
        The penalty is given by the worst miss of the constraint over the entire data column


        :param data: TysonData object
        :return:
        """

        diffarray = np.zeros((len(data[self.target_var]), len(self.target_var_list)))

        for i in range(len(self.target_var_list)):

            self.target_var = self.target_var_list[i]
            self.target_val = self.target_val_list[i]
            self.sign = self.sign_list[i]

            diffs = (data[self.target_var] - self.target_val) * self.weightlist[i]
            if self.sign == '>' or self.sign == '>=':
                diffs = -diffs
            diffarray[:, i] = diffs

        # At each time point, take the best possible result.
        overall_diffs = np.min(diffarray, axis=1)

        # Then take the worst answer over all of the time points
        worstdiff = np.max(overall_diffs)

        return max(0., worstdiff)

    def str_info(self, penalty=None):
        return ' or '.join([self.target_var_list[i]+self.sign_list[i]+str(self.target_val_list[i])
                            for i in range(len(self.target_var_list))]) + ' always'

    def str_info_latex(self, penalty=None):
        return ' or '.join(['$'+self.target_var_list[i] + self.sign_list[i] + str(self.target_val_list[i])+'$'
                            for i in range(len(self.target_var_list))]) + ' always'

    def str_info_latex_mathy(self, penalty=None):
        return latex_escape('$[{:.3g}]$ AlwaysOr constraint you need to format yourself'.format(penalty))


class ParseError(RuntimeError):
    pass


@jit(nopython=True)
def find_first(item, vec):
    """
    return the index of the first occurence of item in vec

    Fast helper function
    """
    for i in range(len(vec)):
        if item == vec[i]:
            return i
    return -1

def latex_escape(s):
    return re.sub('_','\\_',s)

def sign_format(sign, flip=False):
    flips = {'>': '<', '<': '>', '>=': '<', '<=': '>'}
    converts = {'>': '>', '<': '<', '>=': '\\geq', '<=': '\\leq'}
    if sign not in flips:
        raise ValueError('Invalid sign for sign_format: %s' % sign)
    if flip:
        sign = flips[sign]
    return converts[sign]

def float_clean(n):
    if n == int(n):
        return str(int(n))
    return str(n)
