import integrate_caller
import constraint
import numpy as np
import matplotlib.pyplot as plt
import data
import pset
import copy
import multiprocessing
import pickle
from scipy.signal import argrelextrema
from algorithms import ParallelDifferentialEvolution as DiffEvolution, ScatterSearch

PARAM_NAMES = ['gamma', 'gammaki', 'gammacp', 'gammatem', 'sig', 'signet', 'ks_n3', 'Jn3', 'Dn3', 'kd_n3', 'ks_k2',
               'kd_k2', 'kdp_i5', 'kdp_i5_14', 'kp_i5', 'kp_i5_n3', 'kp_i5_k2', 'kp_i5_n2', 'kp_i5_b5', 'kdp_bf',
               'kp_bf_b2', 'ks_n2', 'ks_n2_bf', 'kd_n2', 'ks_ki', 'ks_ki_swi5', 'kd_ki', 'kd_kip', 'kp_ki_e',
               'e_ki_n3', 'e_ki_k2', 'e_ki_n2', 'e_ki_b5', 'e_ki_b2', 'kdp_ki', 'kdp_ki_14', 'ks_b5', 'ks_b5_bf',
               'kd_b5', 'kd_b5_20', 'ks_b2', 'ks_b2_m1', 'kd_b2', 'kd_b2_20', 'kd_b2_h1', 'ks_bud_e', 'e_bud_n3',
               'e_bud_n2', 'e_bud_b5', 'e_bud_b2', 'kd_bud', 'ks_spn', 'kd_spn', 'Jspn', 'ks_ori_e', 'e_ori_b5',
               'e_ori_b2', 'kd_ori', 'ks_swi5', 'ks_swi5_m1', 'kd_swi5', 'ka_swi5_14', 'ki_swi5_b2', 'ka_m1_b2',
               'ki_m1', 'ks_20', 'ks_20_m1', 'kd_20', 'ka_20', 'kd_b5_20_i', 'kd_b2_20_i', 'ki_20_ori', 'ka_cp_b2',
               'ki_cp', 'ka_h1', 'ka_h1_14', 'ki_h1', 'ki_h1_e', 'e_h1_n3', 'e_h1_n2', 'e_h1_b5', 'e_h1_b2', 'kdp_net',
               'kdp_net_14', 'kdp_net_px', 'kp_net', 'kp_net_b2', 'kp_net_en', 'kp_net_15', 'ka_px', 'ki_px',
               'ki_px_p1', 'ks_pds', 'kd_pds', 'kd_pds_20', 'ka_15', 'ka_15_14', 'ki_15', 'ki_15_b2', 'ka_tem',
               'ka_tem_lo', 'ka_tem_p1', 'ki_tem', 'ki_tem_px', 'ks_lo', 'ks_lo_m1', 'kd_lo', 'kd_lo_h1', 'ka_lo',
               'ka_lo_b2', 'ki_lo', 'kas_net', 'f', 'MDT', 'WHI5T', 'SBFT', 'MCM1T', 'APCPT', 'CDH1T', 'NET1T',
               'CDC14T', 'PPXT', 'ESP1T', 'CDC15T', 'TEM1T', 'kd_pds_20_i']
PHASE_PARAMS = ['phialpha', 'phi15', 'phi28']
IC_NAMES = ['V', 'Cln3', 'Bck2', 'WHI5dep', 'SBFdep', 'Cln2', 'CKIT', 'CKIP', 'Clb5T', 'Clb2T', 'BUD', 'ORI', 'SPN',
            'SWI5T', 'CDC20T', 'CDC20A_APCP', 'APCP', 'CDH1A', 'NET1deP', 'PPX', 'PDS1T', 'CDC15', 'TEM1', 'POLOT',
            'POLOA', 'CDC20A_APC']
OUTPUT_NAMES = ['time']+IC_NAMES+['ESP1', 'SBFoverV', 'CLB2CLB5', 'FLAG_UDNA', 'FLAG_BUD', 'FLAG_SPC', 'DIV_COUNT',
                                  'CLB2', 'CLB5']
OUTPUT_HEADERS = {OUTPUT_NAMES[i]: i for i in range(len(OUTPUT_NAMES))}

cbm = ['#009292', '#db6d00', '#b66dff', '#920000', '#24ff24', '#ff6db6']

class FakePool:
    """
    Helper for testing the code that acts like a parallel pool

    """

    def map(self, f, iterable):
        return map(f, iterable)


class Optimizer:

    def __init__(self, con_file, boxsize, exp_file=None, constants=[], log_space=False, spellman=True, ic_path='dat/TL_set.txt'):
        """
        Begins an optimization using the specified files.

        :param con_file: Path to a .con file containing the constraints.
        :param boxsize: Size of the bounding box on each param. In regular space, interpreted as a percent factor, eg 0.2 = +-20%
        In log space, treated as a +- # of factors of 10.
        :param constants: List of variable names that are to be held constant at the Tyson value, not fit.
        :param log_space: Whether the fitting is to be performed in log space
        :param spellman: If True, we are planning to use the Spellman data for fitting, and should maintain the
        appropriate phi parameters.
        :param ic_path: Path to the file that contains the initial conditions for fitting.
        """

        # Returns a dictionary object with keys giving tuples of Cases and values giving a list of Constraints
        self.constraints = constraint.ConstraintParser.parse(con_file)
        self.constraint_list = list(self.constraints.keys())
        self.constants = constants
        self.constant_dict = dict()
        if len(constants) > 0:
            # For anything that was declared constant, look up Tyson's final value
            tyson_data = load_txt_params('dat/Oguz_OPTIMAL.txt')
            self.constant_dict = {p: tyson_data[p] for p in constants}

        self.ic_path = ic_path

        self.log_space = log_space
        self.boxsize = boxsize
        self.total_calls = 0 # track progress of fitting

        self.exp_data = dict()
        self.exp_sigma = dict()

        # Load experimental data
        self.spellman = spellman
        if spellman:
            self.spellman_load()

        if exp_file is not None:
            with open(exp_file) as f:
                for line in f:
                    parts = line.split()
                    self.exp_data[parts[0]] = float(parts[1])
                    self.exp_sigma[parts[0]] = float(parts[2])

    def profile_likelihood(self, variable, values, use_quant=True, use_qual=True, plotit=False, seed=413,
                           mylabel=None, target_func=None):
        """
        Construct a profile likelihood for the specified variable with several consecutive fitting runs.

        :param variable: Which variable are we calculating the profile for?
        :param values: Which values of that varaible are we testing?
        :return:
        """
        np.random.seed(seed)

        if not target_func:
            # Decide which optimization function we are using based on use_quant and use_qual
            if use_quant and use_qual:
                target_func = self.combo_target
                label = 'both'
            elif use_quant:
                target_func = self.quant_target
                label = 'quant'
            elif use_qual:
                target_func = self.qual_target
                label = 'qual'
            else:
                raise ValueError("You can't run profile likelihood using neither quant nor qual data")

        if mylabel:
            label = mylabel

        if type(values) is int:
            # Construct the profile value list based on the box and initial conditions
            ics = load_txt_params('dat/BEST_FIT.txt') # Formerly TL_SET
            print(ics[variable])
            if self.log_space:
                raise NotImplementedError
            else:
                values = np.linspace(ics[variable]*(1-self.boxsize), ics[variable]*(1+self.boxsize), values)

        print(values)

        # Run a simulation at each profile point.
        value_list = []
        score_list = []
        params_list = []
        for val in values:
            self.constant_dict[variable] = val
            score, params = self.run_optimize(target=target_func, iters=2000)
            print('Profile point for %s=%f completed with score %f' % (variable, val, score))
            value_list.append(val)
            score_list.append(score)
            params_list.append(params)

        f = open('results/profile_%s_%s_%i' % (label, variable, seed), 'wb')
        pickle.dump((value_list, score_list, params_list), f)
        f.close()
        del self.constant_dict[variable]

        if plotit:
            plt.figure()
            plt.plot(value_list, score_list, 'o-')
            plt.xlabel(variable)
            plt.ylabel('Objective')

    def profile_range(self, param, number):
        """Generates a list of values for profile likelihood.
        Starting with the TLset value, moves by factors of sqrt(10)"""
        ics = load_txt_params(self.ic_path)
        ic = ics[param]
        values = [ic*(10**(n/2)) for n in range(-number, number+1)]
        return values

    def basic_sensitivity(self, paramset, param, values=None, target=None):
        """ Checks how the objective function changes if I move this param but keep everything else constant"""
        if not target:
            target = self.all_3_target
        if not values:
            values = np.linspace(paramset[param]-2., paramset[param]+2., 17)

        results = []
        for v in values:
            new_dict = copy.copy(paramset._param_dict)
            new_dict[param] = v
            new_pset = pset.PSet(new_dict, allow_negative=True)
            dict_to_score = self.pset_to_paramdict(new_pset)
            results.append(target(dict_to_score))

        return values, results


    def run_optimize_quant(self, iters=500):

        print("Starting optimization")

        variables = self.prepare_variables()

        alg = DiffEvolution(variables, mutation_rate=0.5, mutation_factor=1.0, num_per_island=40, num_islands=1,
                            migrate_every=1e9, max_iterations=iters, num_to_migrate=0)
        joblist = alg.start_run()
        print('Initial job queue of size %i created' % len(joblist))

        while len(joblist) > 0:
            formatted_joblist = [self.pset_to_paramdict(j) for j in joblist]
            results = parallel_pool.map(self.quant_target, formatted_joblist)
            new_joblist = []
            for result, job in zip(results, joblist):
                new_joblist += alg.got_result(job, result)
            joblist = new_joblist

        best_index = np.argmin(alg.fitnesses[0])
        return alg.fitnesses[0][best_index], alg.individuals[0][best_index]

    def run_optimize(self, target, iters=500, algorithm='de'):
        """Runs an optimization on whichever target was specified, e.g. self.quant_target or self.combo_target"""

        print("Starting optimization")

        variables = self.prepare_variables()

        if algorithm=='ss':
            alg = ScatterSearch(variables, popsize=12, maxiters=iters, saveevery=500)
        elif algorithm == 'de':
            alg = DiffEvolution(variables, num_per_island=80, num_islands=1, max_iterations=iters, save_every=100,
                                mutation_rate=0.5, mutation_factor=1.0, migrate_every=1e9, num_to_migrate=0)
        elif algorithm == 'dei':
            alg = DiffEvolution(variables, num_per_island=80, num_islands=5, max_iterations=iters, save_every=100,
                                mutation_rate=0.5, mutation_factor=1.0, migrate_every=500, num_to_migrate=10)
        else:
            raise ValueError("Invalid algorithm %s" % algorithm)

        joblist = alg.start_run()
        print('Initial job queue of size %i created' % len(joblist))

        done = False
        while len(joblist) > 0 and not done:
            formatted_joblist = [self.pset_to_paramdict(j) for j in joblist]
            results = parallel_pool.map(target, formatted_joblist)
            new_joblist = []
            for result, job in zip(results, joblist):
                new_jobs = alg.got_result(job, result)
                if new_jobs == 'STOP':
                    done = True
                    break
                new_joblist += new_jobs
            joblist = new_joblist

        if algorithm == 'de':
            best_index = np.argmin(alg.fitnesses[0])
            return alg.fitnesses[0][best_index], alg.individuals[0][best_index]
        elif algorithm == 'dei':
            best_index = np.argmin(alg.fitnesses)
            best_i = int(best_index / 80)
            best_j = best_index % 80
            return alg.fitnesses[best_i][best_j], alg.individuals[best_i][best_j]
        elif algorithm == 'ss':
            # Scatter
            finals = alg.refs + alg.local_mins
            best = min(finals, key=lambda x: x[1])
            return best[1], best[0]


    def prepare_variables(self):
        """ Prepares the list of variables and bounds to pass to the algorithm"""
        ics = load_txt_params(self.ic_path)
        var_list = []
        for v in ics:
            if v in self.constant_dict:
                continue
            if self.log_space:
                if ics[v]==0:
                    # Bad... just put some arbirary low value on it
                    var_list.append((v, -8., -10.))
                else:
                    var_list.append((v, np.log10(ics[v])-self.boxsize, np.log10(ics[v])+self.boxsize))
            else:
                var_list.append((v, ics[v]*(1-self.boxsize), ics[v]*(1+self.boxsize)))
        if self.spellman:
            for v in PHASE_PARAMS:
                var_list.append((v, 500., 650.))
        return var_list


    def pset_to_paramdict(self, pset):
        """
        Converts a pset to a dict to get passed to the target function

        :param pset:
        :return:
        """
        if self.log_space:
            param_dict = {p:10.**pset._param_dict[p] for p in pset._param_dict if p not in PHASE_PARAMS}
            if self.spellman:
                # Un-log the phase parameters
                param_dict = {**param_dict, **{v: pset[v] for v in PHASE_PARAMS}}
            return {**param_dict, **self.constant_dict}
        else:
            return {**copy.copy(pset._param_dict), **self.constant_dict}


    @staticmethod
    def target_piecewise(params, self):
        """
        Returns a dictionary with keys giving tuples of Cases and values giving the penalty value associated with
        that case

        :param params: dictionary of parameter values
        :param self: Optmizer object
        :return: dict
        """
        self.total_calls += 1
        results = list(map(Optimizer.run_case,
                                         [(copy.copy(params), casetuple, self.constraints[casetuple]) for
                                         casetuple in self.constraint_list]))

        return {self.constraint_list[i]: results[i] for i in range(len(self.constraint_list))}

    @staticmethod
    def target_extra_piecewise(params, self):
        """
        Returns a dictionary with keys giving tuples of Cases and values giving a dictionary mapping constraints to
        penalty values.

        :param params: dictionary of parameter values
        :param self: Optmizer object
        :return: dict
        """
        self.total_calls += 1
        results = list(map(Optimizer.run_case_piecewise,
                                         [(copy.copy(params), casetuple, self.constraints[casetuple]) for
                                         casetuple in self.constraint_list]))

        return {self.constraint_list[i]: results[i] for i in range(len(self.constraint_list))}

    @staticmethod
    def target_full(params, self):
        """
        Returns the total penalty function value
        :param params: dictionary of parameter values
        :param self: Optimizer object
        :return:
        """
        return sum(Optimizer.target_piecewise(params, self).values())

    @staticmethod
    def run_case(caseparams):
        """
        Function to be called by multiprocessing.map() that runs one case

        :param caseparams: An ordered pair (params, casetuple, constraintlist).
        self is the Optimizer running this optimization
        params is a dict consisting of the unaltered parameter values and initial conditions.
        casetuple is a tuple of Case objects indicating how the params should be modified for this case.
        constraintlist is a list of constraints to be satisfied for this case.
        get_data (optional) is boolean. If true, return the Data object instead of the
        :return: The total value of the penalty function for this case.
        """
        (params, casetuple, constraintlist) = caseparams

        # Update params as determined by the cases
        for case in casetuple:
            exec(case.funcstr)

        # Call the C++ code
        paramarray, icarray = input_dict_to_arrays(params)
        resultarray = integrate_caller.call(paramarray, icarray)

        # Convert the array to a Data object\
        resultdata = data.Data()
        resultdata.data = resultarray
        resultdata.cols = OUTPUT_HEADERS

        penalties = [c.penalty(resultdata) for c in constraintlist]
        return sum(penalties)

    @staticmethod
    def run_case_piecewise(caseparams):
        """
        run_case version where, within the case, shows the penalty for each constraint.
        """
        (params, casetuple, constraintlist) = caseparams

        # Update params as determined by the cases
        for case in casetuple:
            exec(case.funcstr)

        # Call the C++ code
        paramarray, icarray = input_dict_to_arrays(params)
        resultarray = integrate_caller.call(paramarray, icarray)

        # Convert the array to a Data object\
        resultdata = data.Data()
        resultdata.data = resultarray
        resultdata.cols = OUTPUT_HEADERS

        penalties = {c: c.penalty(resultdata) for c in constraintlist}
        return penalties

    @staticmethod
    def quant_extract(params):
        """
        Runs simulations and extracts the quantitative values that correspond to the known experimental data.

        :param params: dict consisting of the unaltered parameter values and initial conditions.
        :return:
        """

        sim_data = dict()

        params['MDT'] = 90

        for name, f in [('D', 0.4), ('M', 0.6)]:
            params['f'] = f
            # Call the C++ code
            paramarray, icarray = input_dict_to_arrays(params)
            dataarray = integrate_caller.call(paramarray, icarray)

            dt = dataarray[1, 0]  # time step
            # Finds all local maxima of V, which are cell divisions.
            # [0] index is because return is a 1-tuple containing answer
            daughter_divs = argrelextrema(dataarray[:, 1], np.greater)[0]

            # If there are too few divisions, count from the last div to the simulation end as one cycle
            # This will happen if cell is inviable, and will give a very bad score
            if len(daughter_divs) == 0:
                daughter_divs = np.array([0, dataarray.shape[0]-1])
            elif len(daughter_divs) < 3:
                daughter_divs = np.array(list(daughter_divs) + [dataarray.shape[0]-1])

            # Apply the problem-specific check of consistent volume at cell division (last 3 divisions)
            vols_at_div = dataarray[daughter_divs[-3:], 1]
            sim_data['%s_Vratio' % name] = max(vols_at_div) / min(vols_at_div)

        for field in ['Vratio']:
            sim_data[field] = (sim_data['D_%s' % field] + sim_data['M_%s' % field]) / 2.

        return sim_data

    def quant_target(self, params):
        obj = 0.

        sim_data = Optimizer.quant_extract(params)
        for p in self.exp_data:
            obj += 1./self.exp_sigma[p]**2 * (sim_data[p]-self.exp_data[p])**2

        obj = min(obj, 1e12) # In case awful overflow things happend?

        return obj

    def combo_target(self, params):
        """
        Combines quantitative and qualitative data with a static penalty function

        :param params:
        :return:
        """

        return self.quant_target(params) + Optimizer.target_full(params, self)

    def spellman_load(self):
        """
        Load the Spellman 1998 data files. Must call this before using spellman_target

        :param params:
        :return:
        """
        self.spellman_alpha = data.Data()
        self.spellman_alpha.load_data('dat/alpha.exp')
        self.spellman_cdc15 = data.Data()
        self.spellman_cdc15.load_data('dat/cdc15.exp')
        self.spellman_cdc28 = data.Data()
        self.spellman_cdc28.load_data('dat/cdc28.exp')


    def spellman_target(self, params):
        """
        Calculates the target function based on the time courses from Spellman 1998

        :param params:
        :return:
        """

        # The start of the simulation is often wonky, so let's instead play by the following rules:
        # Each experiment has a phase offset parameter phi from 100 to 200, to allow it to align to the correct
        # phase of the cycle at some point after the initial equilibration.

        datasets = ((self.spellman_alpha, 66., params['phialpha']), (self.spellman_cdc15, 70., params['phi15']),
                    (self.spellman_cdc28, 90., params['phi28']))

        total_error = 0.
        for expdat, mdt, offset in datasets:
            params['MDT'] = mdt
            # Call the C++ code
            paramarray, icarray = input_dict_to_arrays(params)
            dataarray = integrate_caller.call(paramarray, icarray)

            # adjust times according to the phase offset
            dataarray[:, 0] -= offset

            # Figure out which dataarray rows we are using
            rowindices = []
            exppointer = 0
            datapointer = 0
            while exppointer < expdat.data.shape[0]:
                datapointer += 1
                if dataarray[datapointer, 0] >= expdat['Time'][exppointer]:
                    rowindices.append(datapointer)
                    exppointer += 1

            # Iterate through the variables for which we have data
            for var in expdat.cols:
                if var == 'Time':
                    continue

                # Normalize the same way as the experimental data is normalized
                sim_col = dataarray[:, OUTPUT_NAMES.index(var)]
                sim_col /= np.max(sim_col)
                if np.any(np.isnan(sim_col)):
                    # This simulation really blew it and generated nans! Give an arbitrary high penalty.
                    total_error += 1000
                    continue

                # Alt normalization into the weird log space used in Spellman 1998
                # sim_col = np.log2(sim_col) #This is bad - simulation regularly outputs 0's.
                # sim_col -= np.average(sim_col)

                # Inject custom penalty for having too many peaks (you're oscillating really fast to hit all the points)
                maxs = argrelextrema(sim_col, np.greater)[0]
                total_error += max(0, len(maxs)-20)

                exp_col = expdat[var]

                # Add in the error at each point
                for i in range(len(exp_col)):
                    if np.isnan(exp_col[i]):
                        continue
                    total_error += (sim_col[rowindices[i]] - exp_col[i]) ** 2.

        return total_error

    def spellman_quant_target(self, params):
        """Target function using both the Spellman data and the regular quant data (the latter could be used, e.g
        to just enforce the consistent volume rule."""
        return self.spellman_target(params) + self.quant_target(params)

    def all_3_target(self, params):
        """ Combines the Spellman, quant, and qual target functions"""
        # Arbitrary 0.1 multiplier on the qual data to make it of roughly the same magnitude as the spellman data.
        return self.spellman_target(params) + self.quant_target(params) + 0.0667*Optimizer.target_full(params, self)

    def qual_target(self, params):
        return Optimizer.target_full(params, self)

    def tyson_original_target(self, params):
        """
        Returns a dictionary with keys giving tuples of Cases and values giving the penalty value associated with
        that case

        :param params: dictionary of parameter values
        :param self: Optmizer object
        :return: dict
        """

        # Note: if you ever care to use this target for optimization, you should not load this thing every time!
        skeleton_cases = constraint.ConstraintParser.parse_skeleton('dat/tyson13-skeleton.con')
        params['f'] = 0.4
        params['MDT'] = 100.
        original_params = copy.copy(params)
        objective = 0.

        for casetuple in skeleton_cases:
            params = copy.copy(original_params)

            # Update params as determined by the cases
            for case in casetuple:
                exec(case.funcstr)

            # Call the C++ code
            paramarray, icarray = input_dict_to_arrays(params)
            resultarray = integrate_caller.call2000(paramarray, icarray) #  Use the sim out to time 2000, same as Oguz 2013

            # Convert the array to a Data object\
            resultdata = data.Data()
            resultdata.data = resultarray
            resultdata.cols = OUTPUT_HEADERS

            answer = skeleton_cases[casetuple][1]
            if answer!='V':
                answer = 'I'

            if max(resultdata['V']) > 25:
                simresult = 'I'
            else:
                # Finds all local maxima of V, which are cell divisions.
                # [0] index is because return is a 1-tuple containing answer
                daughter_divs = argrelextrema(resultdata['V'], np.greater)[0]
                if len(daughter_divs) < 3:
                    simresult = 'N'
                elif 0.95 < resultdata['V'][daughter_divs[-1]] / resultdata['V'][daughter_divs[-2]] < 1.05 and 0.95 < resultdata['V'][daughter_divs[-1]] / resultdata['V'][daughter_divs[-3]] < 1.05:
                    simresult = 'V'
                else:
                    simresult = 'N'
            if answer == simresult:
                objective -= 1.
        return objective

# The following functions are provided to aid in visualization and analysis of the fitting results.
# For example usage, see production_sandbox.py

def input_dict_to_arrays(paramdict):
    param = np.array([paramdict[p] for p in PARAM_NAMES])
    ic = np.array([paramdict[p] for p in IC_NAMES])
    return param, ic


def load_txt_params(file='dat/Oguz_OPTIMAL.txt'):
    """Loads a parameter file of the same format as Oguz_OPTIMAL.txt as a dictionary"""
    result = dict()
    varnames = PARAM_NAMES + IC_NAMES
    with open(file,'r') as f:
        i = 0
        for line in f:
            result[varnames[i]] = float(line)
            i += 1
    return result


def run_case_by_str(string, params=None, toplot=[], log_space=False, min_lines=[], max_lines=[],
                    stage_lines=[], oneaxis=False, special_vertical_line=False, size=(4,4)):
    """
    Runs the simulation with the specified parameters, reading the 'string' input as an instruction for how the
    parameters should be mutataed.

    Optionally plots one or more variables with configurable settings, to generate plots as in Fig. 3, Fig. 4A-C, and
    Fig. S3

    :param string: A string describing how to mutate the parameters. For example, "Dn3=0 Cln3=0" could be used to
    simulate a cln3 knockout mutant.
    :type string: str
    :param params: Parameter set to use
    :type params: pset.PSet
    :param toplot: List of output variables to plot
    :type toplot: list of str
    :param min_lines: Dict of parameter_name:lower_constraint, draws a horizontal line to indicate that it was
    supposed to reach at least here.
    :param max_lines: Dict of parameter_name:upper_constraint, draws a horizontal line to indicate that it was
    not supposed to reach here.
    :param stage_lines: List of params for which I should draw vertical lines indicating the completion of each cell
    cycle stage
    :param oneaxis: If true, plot all the variables on one axis.
    :param special_vertical_line: Single special case: Draws a vertical black line where V hits 8.
    :return:
    """
    if params is None:
        params = load_txt_params()
    else:
        params = copy.copy(params._param_dict)

    if log_space:
        params = {p:10.**params[p] for p in params if p not in PHASE_PARAMS}

    params['MDT'] = 100.
    params['f'] = 0.4

    # Parse and apply the case
    casetuple = tuple([constraint.Case(casestr) for casestr in string.split()])
    for case in casetuple:
        exec(case.funcstr)

    # Call the C++ code
    paramarray, icarray = input_dict_to_arrays(params)
    resultarray = integrate_caller.call(paramarray, icarray)

    # Convert the array to a Data object\
    resultdata = data.Data()
    resultdata.data = resultarray
    resultdata.cols = OUTPUT_HEADERS

    # Plot the results
    figs = []  # So we can return the handles
    plt.rc('font', **{'size': 16})
    if oneaxis:
        f = plt.figure(figsize=size).number
        figs = [f]
    plotnum = 0
    for p in toplot:
        plotnum += 1
        if not oneaxis:
            f = plt.figure(figsize=size).number
        if not oneaxis or plotnum == 1:
            plt.plot(resultdata['time'], resultdata[p], color=cbm[0])
        else:
            plt.plot(resultdata['time'], resultdata[p], color=cbm[1], dashes=(6,4))
        if p in min_lines:
            plt.axhline(y=min_lines[p], color='k', dashes=[8, 4, 4, 4])
        if p in max_lines:
            plt.axhline(y=max_lines[p], color=cbm[3], dashes=[4, 4])
            pass
        if p in stage_lines:
            cm = plt.get_cmap('Oranges')
            for xx, clr, dsh in zip(['FLAG_UDNA', 'FLAG_SPC', 'DIV_COUNT', 'FLAG_BUD'],
                                    [cm(0.5), cm(0.7), cm(0.9), '#777777'],
                                    [(), (), (), (4, 4)]):
                dat = resultdata[xx]
                # indices where output xx jumps in value (index [0] is b/c return is a 1-tuple containing the answer)
                indices = np.nonzero(dat[1:] > dat[:-1])[0]
                print(indices)
                for i in indices:
                    plt.axvline(x=resultdata['time'][i], color=clr, dashes=dsh, linewidth=3)
            # Origin relicensing
            # indices = np.nonzero(resultdata['ORI'] == 0) # Should work, but generates cryptic bug in plot function.
            indices = []
            for i in range(len(resultdata['ORI'])):
                if resultdata['ORI'][i] == 0:
                    indices.append(i)
            print(indices)
            for i in indices:
                plt.axvline(x=resultdata['time'][i], color=cm(0.3), linewidth=3)
        if special_vertical_line:
            keyindex = np.argmax(resultdata['V'] > 8)
            plt.axvline(x=resultdata['time'][keyindex], color='k')
        if not oneaxis:
            plt.ylabel(p)
        plt.xlabel('time (min)')
        plt.tight_layout()
        if not oneaxis:
            figs.append(f)
    return figs


def plot_spellman_compare(params, params2, toplot, which='alpha'):
    """
    Make graphs that plot time series generated by 2 parameter sets against the quantitative data from Spellman et al.
    1998
    It is suggested you access this by running through compare_to_tyson() which preprocesses the input as needed

    :param params: First parameter set
    :type params: dict
    :param params2: Second parameter set
    :type params2: dict
    :param toplot: Names of variables to plot
    :type toplot: iterable of str
    :param which: Which of the Spellman datasets to use, 'alpha', '15', 'or '28'
    :return:
    """
    # In this one, assumed to input a non-log dict, as in compare_to_tyson()

    # Choose which spellman data set
    choices = {'alpha': ('dat/alpha.exp', 66), '15': ('dat/cdc15.exp', 70), '28': ('dat/cdc28.exp', 90)}
    (file, mdt) = choices[which]

    params['f'] = 0.4
    params['MDT'] = mdt

    spell_data = data.Data()
    spell_data.load_data(file)

    # Call the C++ code
    paramarray, icarray = input_dict_to_arrays(params)
    resultarray = integrate_caller.call(paramarray, icarray)

    # Convert the array to a Data object\
    resultdata = data.Data()
    resultdata.data = resultarray
    resultdata.cols = OUTPUT_HEADERS

    # Repeat for params #2
    # Call the C++ code
    paramarray2, icarray2 = input_dict_to_arrays(params2)
    resultarray2 = integrate_caller.call(paramarray2, icarray2)

    # Convert the array to a Data object\
    resultdata2 = data.Data()
    resultdata2.data = resultarray2
    resultdata2.cols = OUTPUT_HEADERS

    # Plot the results
    plt.rc('font', **{'size': 16})
    for p in toplot:
        plt.figure(figsize=(4, 4))
        plt.rc('font', **{'size': 16})
        ydata = resultdata[p]
        ydata /= max(ydata)  # Normalize to same scale as Spellman data
        plt.plot(resultdata['time'], resultdata[p], '-', color=cbm[0])

        ydata2 = resultdata2[p]
        ydata2 /= max(ydata2)
        myshift = params['phi'+which] - params2['phi'+which]
        plt.plot(resultdata2['time']+myshift, resultdata2[p], '--', color=cbm[1])

        spell_xdata = copy.copy(spell_data['Time'])
        spell_xdata += params['phi'+which]  # Apply time offset
        spell_ydata = spell_data[p]
        plt.plot(spell_xdata, spell_ydata, 'd', color=cbm[3])
        plt.xlim([min(spell_xdata) - 0.25 * (max(spell_xdata) - min(spell_xdata)),
                  max(spell_xdata) + 0.25 * (max(spell_xdata) - min(spell_xdata))])
        plt.ylabel(p)
        plt.xlabel('time (min)')
        plt.tight_layout()


def plot_profiles_triple(param, seeds, logspace=True, stems= None, norm_to_zero=True, use_min=True):
    """
    Make profile likelihood plots similar to Fig. 5

    :param param: Name of the parameter
    :type param: str
    :param seeds: The random seeds that were used in the runs, assumed to be the end of the file names
    :type seeds: iterable
    :param logspace: Was the profile calculated in log10 space
    :type logspace: bool
    :param stems: 3-tuple giving the file names of the quantitative, qualitative, and combined profiles, respectively,
    not including the seed number at the end
    :type stems: tuple of str
    :param norm_to_zero: Should all profiles be shifted down so the min value is zero
    :type norm_to_zero: bool
    :param use_min: Should we plot the minimum value over the n replicates? (If false, instead plot the median with
    error bars)
    :type use_min: bool
    :return:
    """
    if not stems:
        stems = ['results/profiles_final/%s/profile_%s_%s_' % (param, x, param) for x in ('spell','qual','spellnqual')]

    plt.figure(figsize=(4, 3))
    for (stem, line, scale, color) in zip(stems, ('o-','s-','^-'), (1., 15., 1.), cbm[0:3]):
        profiles = []
        vals = None
        for s in seeds:
            f = open(stem + str(s), 'rb')
            res = pickle.load(f)
            f.close()
            if vals is None:
                vals = res[0]
            this_profile = [x/scale for x in res[1]]
            profiles.append(this_profile)

        prof_array = np.array(profiles)
        plt.rc('font', **{'size': 16})
        if use_min:
            dat = np.min(prof_array, axis=0)
            if norm_to_zero:
                dat = dat - np.min(dat)
            plt.plot(vals, dat, line, color=color)
        else:
            med = np.median(prof_array, axis=0)
            if norm_to_zero:
                norm = np.min(med)
            else:
                norm = 0.
            dmin = med - np.min(prof_array, axis=0)
            dmax = np.max(prof_array, axis=0) - med
            med -= norm
            plt.errorbar(vals, med, yerr=np.array([dmin, dmax]), fmt=line, capsize=4, color=color)

    mid = int(prof_array.shape[1] / 2.)  # Middle of list = point of best fit, to draw the line
    plt.axvline(x=vals[mid], color=cbm[3])

    if logspace:
        plt.xscale('log')
    plt.ylim((0,300))
    plt.ylabel('Objective')
    plt.xlabel(param)
    plt.tight_layout()


def compare_to_tyson(your_pset, your_logspace, tyson_phasevals=(501., 587., 558.),
                     quant_to_plot=[], which='alpha'):
    """
    Makes several graphs that compare the given parameter set to the best fit presented in Oguz et al. 2013, which is
    saved in dat/Oguz_OPTIMAL.txt

    Makes a graph that compares the parameter sets in terms of the number of consistent phenotypes, as in Fig. S3

    Makes graphs that plot time series based on the two parameter sets in comparison to the quantitative data from
    Spellman et al. 1998, as in Fig. 4 D-F and Fig. S1

    :param your_pset: Parameter set to use
    :type your_pset: pset.PSet
    :param your_logspace: Whether the parameter set is given in log10 space
    :type your_logspace: bool
    :param tyson_phasevals: A 3-tuple giving the values of phi_alpha, phi_15, phi_28 to use for the Oguz et al.
    parameter set. The default values are optimized to give the best fit to the data
    :type tyson_phasevals: tuple
    :param quant_to_plot: Names of the quantitative variables for which we should make blots
    :type quant_to_plot: iterable
    :param which: Which of the 3 datasets from Spellman et al. should we use, 'alpha', '15', or '28'
    :type which: str
    :return:
    """
    tysonpdict = load_txt_params()
    myopt = Optimizer('dat/tyson13-PW2.con', 2, 'dat/ratio_only.exp', ['MDT', 'f'], your_logspace, spellman=True)
    for i in range(len(PHASE_PARAMS)):
        tysonpdict[PHASE_PARAMS[i]] = tyson_phasevals[i]
    yourpdict = myopt.pset_to_paramdict(your_pset)

    # Comare on qualitative data
    tyson_extra_pieces = Optimizer.target_extra_piecewise(tysonpdict, myopt)
    my_extra_pieces = Optimizer.target_extra_piecewise(yourpdict, myopt)
    tyson_pieces = {c: sum(tyson_extra_pieces[c].values()) for c in tyson_extra_pieces}
    my_pieces = {c: sum(my_extra_pieces[c].values()) for c in my_extra_pieces}
    print('Tyson qual score: '+str(sum(list(tyson_pieces.values()))))
    print('Your qual score: ' + str(sum(list(my_pieces.values()))))

    tyson_threshold_scores = [len([x for x in tyson_pieces.values() if x <= thr]) for thr in range(11)]
    my_threshold_scores = [len([x for x in my_pieces.values() if x <= thr]) for thr in range(11)]
    plt.figure(figsize=(4,4))
    plt.rc('font', **{'size': 16})
    plt.plot(range(11), my_threshold_scores, 'o-', label='This study', color=cbm[0])
    plt.plot(range(11), tyson_threshold_scores, 'v--', label='Oguz 2013', color=cbm[1])
    plt.xlabel('Threshold')
    plt.ylabel('# Correct')
    plt.tight_layout()
    plt.legend()

    # Compare on quantitative data
    tyson_quant = myopt.spellman_target(tysonpdict)
    your_quant = myopt.spellman_target(yourpdict)
    print('Tyson quant score: '+str(tyson_quant))
    print('Your quant score: '+str(your_quant))

    # Compare on tyson's objective function
    tyson_tyson = myopt.tyson_original_target(tysonpdict)
    your_tyson = myopt.tyson_original_target(yourpdict)
    print('Tyson Tyson score: '+str(tyson_tyson))
    print('Your Tyson score: '+str(your_tyson))

    plot_spellman_compare(yourpdict, tysonpdict, quant_to_plot, which=which)


def params_to_txt(params, outname, logspace=True, const=('f', 'MDT')):
    """
    Write the parameter set to a file, in the same format as BEST_FIT.txt and TL_set.txt
    This file format is needed to set up profile likelihood runs.

    :param params: Parameter set
    :type params: pset.PSet
    :param outname: File name where we should save the output
    :type outname: str
    :param log_space: Whether values in the PSet are given in log10 space
    :type log_space: bool
    :param: const: Parameters that were held constant during fitting, and therefore are not contained in the PSet
    object. For these parameters, we should instead just use the value in dat/TL_set.txt
    :type const: Iterable
    """
    out = open(outname, 'w')
    ics = load_txt_params('dat/TL_set.txt')
    for p in PARAM_NAMES+IC_NAMES:
        if p in const:
            out.write('%g\n' % ics[p])
        else:
            if logspace:
                out.write('%g\n' % 10.**params[p])
    out.close()


def run_sim_to_file(params, filename, log_space=True):
    """
    Run the cell cycle model using the given parameters, and save the results to a text file

    :param params: Parameter set
    :type params: pset.PSet
    :param filename: File name where we should save the results
    :type filename: str
    :param log_space: Whether values in the PSet are given in log10 space
    :type log_space: bool
    :return:
    """
    p_dict = {k: params[k] for k in params.keys()}
    if log_space:
        p_dict = {p:10.**p_dict[p] for p in p_dict if p not in PHASE_PARAMS}

    p_dict['MDT'] = 100
    p_dict['f'] = 0.4
    paramarray, icarray = input_dict_to_arrays(p_dict)
    resultarray = integrate_caller.call(paramarray, icarray)

    f = open(filename, 'w')
    f.write('#')
    f.write('\t'.join(OUTPUT_NAMES))
    for i in range(resultarray.shape[0]):
        f.write('\n')
        f.write('\t'.join([str(x) for x in list(resultarray[i,:])]))
    f.close()


if __name__ == '__main__':
    # Perform the optimization runs. Results are saved in binary pickle format.

    #  Perform parameter fitting of the model.
    parallel_pool = multiprocessing.Pool(40)
    myopt = Optimizer('dat/tyson13-PW2.con', 2, 'dat/ratio_only.exp', ['MDT', 'f'], True, spellman=True)
    np.random.seed(1123)
    result = myopt.run_optimize(myopt.all_3_target, 70000, algorithm='ss')
    print(result)
    f = open('final_fit', 'wb')
    pickle.dump(result, f)
    f.close()

    #  Perform profile likelihood analysis on selected parameters (uncomment this section to run)
    # myopt = Optimizer('dat/tyson13-PW2.con', 0.5, 'dat/ratio_only.exp', ['MDT', 'f'], False, spellman=True, ic_path='dat/BEST_FIT.txt')
    # my_q_opt = Optimizer('dat/tyson13-PW2.con', 0.5, 'dat/ratio_only.exp', ['MDT', 'f'], False, spellman=False, ic_path='dat/BEST_FIT.txt')
    #
    # for v in ['ks_spn', 'ESP1T', 'kdp_i5', 'ki_m1']:
    #     for s in [42, 43, 44, 45, 46]:
    #         myopt.profile_likelihood(v, myopt.profile_range(v, 4), seed=s, mylabel='spell',
    #                                   target_func=myopt.spellman_quant_target)
    #         myopt.profile_likelihood(v, myopt.profile_range(v, 4), seed=s, mylabel='spellnqual',
    #                              target_func=myopt.all_3_target)
    #         my_q_opt.profile_likelihood(v, myopt.profile_range(v, 4), seed=s, mylabel='qual',
    #                              target_func=myopt.qual_target)

    parallel_pool.close()
else:
    parallel_pool = FakePool()
